import freeOrionAIInterface as fo
from logging import debug
from typing import NamedTuple, Optional

import AIDependencies
import EspionageAI
import MilitaryAI
import PlanetUtilsAI
from aistate_interface import get_aistate
from buildings import BuildingType
from colonization import get_nest_rating, special_is_nest
from colonization.calculate_influence import calculate_influence
from colonization.calculate_population import calc_max_pop
from colonization.calculate_production import calculate_production
from colonization.calculate_research import calculate_research
from colonization.calculate_stability import calculate_stability
from colonization.claimed_stars import count_claimed_stars
from colonization.colony_score import MINIMUM_COLONY_SCORE, RESOURCE_PRIORITY_MULTIPLIER
from colonization.rate_pilots import detection_value, rate_colony_for_pilots
from common.fo_typing import PlanetId, SpeciesName, SystemId
from empire.growth_specials import get_growth_specials
from EnumsAI import FocusType, MissionType, PriorityType
from freeorion_tools import (
    assertion_fails,
    get_named_real,
    get_partial_visibility_turn,
    get_species_supply,
    tech_is_complete,
    tech_soon_available,
)
from freeorion_tools.caching import cache_for_current_turn
from PlanetUtilsAI import get_empire_populated_planets, stability_with_focus
from turn_state import (
    get_empire_outposts,
    get_empire_planets_by_species,
    get_owned_planets_in_system,
    get_supply_group,
    get_supply_group_id,
    get_system_supply,
    have_computronium,
    have_honeycomb,
    have_worldtree,
    owned_asteroid_coatings,
    supply_connected,
)
from universe.system_network import get_neighbors, within_n_jumps

empire_metabolisms = {}


def calculate_planet_colonization_rating(
    *,
    planet_id: PlanetId,
    mission_type: MissionType,
    spec_name: Optional[SpeciesName],
    detail: Optional[list],
) -> float:
    """Returns the colonisation value of a planet."""
    if spec_name is None:
        spec_name = ""

    planet = fo.getUniverse().getPlanet(planet_id)
    rating, new_detail = _calculate_planet_rating(planet, spec_name)
    owned = int(planet.owner == fo.empireID())
    debug(f"colrating {mission_type}-{owned} {planet} {spec_name} rating={rating}, detail={new_detail}")
    if detail:
        detail.clear()
        detail.extend(new_detail)
    return rating


def _determine_colony_threat_factor(planet, spec_name, existing_presence):
    sys_status = get_aistate().systemStatus.get(planet.systemID, {})
    cur_best_mil_ship_rating = max(MilitaryAI.cur_best_mil_ship_rating(), 0.001)
    local_defenses = sys_status.get("all_local_defenses", 0)
    local_threat = sys_status.get("fleetThreat", 0) + sys_status.get("monsterThreat", 0)
    neighbor_threat = sys_status.get("neighborThreat", 0)
    jump2_threat = 0.6 * max(0, sys_status.get("jump2_threat", 0) - sys_status.get("my_neighbor_rating", 0))
    area_threat = neighbor_threat + jump2_threat
    area_threat *= 2.0 / (existing_presence + 2)  # once we have a foothold be less scared off by area threats
    # TODO: evaluate detectability by specific source of area threat, also consider if the subject AI already has
    # researched techs that would grant a stealth bonus
    local_enemies = sys_status.get("enemies_nearly_supplied", [])
    # could more conservatively base detection on get_aistate().misc.get("observed_empires")
    if not EspionageAI.colony_detectable_by_empire(planet.id, spec_name, empire=local_enemies, default_result=True):
        area_threat *= 0.05
    net_threat = max(0, local_threat + area_threat - local_defenses)
    # even if our military has lost all warships, rate planets as if we have at least one
    reference_rating = max(
        cur_best_mil_ship_rating,
        MilitaryAI.get_preferred_max_military_portion_for_single_battle()
        * MilitaryAI.get_concentrated_tot_mil_rating(),
    )
    threat_factor = min(1.0, reference_rating / (net_threat + 0.001)) ** 2
    if threat_factor < 0.5:
        mil_ref_string = "Military rating reference: %.1f" % reference_rating
        debug(
            "Significant threat discounting %2d%% at %s, local defense: %.1f, local threat %.1f, area threat %.1f"
            % (100 * (1 - threat_factor), planet.name, local_defenses, local_threat, area_threat)
            + mil_ref_string
        )
    return threat_factor


def _revise_threat_factor(
    threat_factor: float,
    planet_value: float,
    system_id: int,
    min_planet_value: float = MINIMUM_COLONY_SCORE,
) -> float:
    """
    Check if the threat_factor should be made less severe.

    If the AI does have enough total military to secure this system, and the target has more than minimal value,
    don't let the threat_factor discount the adjusted value below min_planet_value +1, so that if there are no
    other targets the AI could still pursue this one.  Otherwise, scoring pressure from
    MilitaryAI.get_preferred_max_military_portion_for_single_battle might prevent the AI from pursuing a heavily
    defended but still obtainable target even if it has no softer locations available.

    :param threat_factor: the base threat factor
    :param planet_value: the planet score
    :param system_id: the system ID of subject planet
    :param min_planet_value: a floor planet value if the AI has enough military to secure the system
    :return: the (potentially) revised threat_factor
    """

    # the default value below for fleetThreat shouldn't come in to play, but is just to be absolutely sure we don't
    # send colony ships into some system for which we have not evaluated fleetThreat
    system_status = get_aistate().systemStatus.get(system_id, {})
    system_fleet_treat = system_status.get("fleetThreat", 1000)
    # TODO: consider taking area threat into account here.  Arguments can be made both ways, see discussion in PR2069
    sys_total_threat = system_fleet_treat + system_status.get("monsterThreat", 0) + system_status.get("planetThreat", 0)
    if (MilitaryAI.get_concentrated_tot_mil_rating() > sys_total_threat) and (planet_value > 2 * min_planet_value):
        threat_factor = max(threat_factor, (min_planet_value + 1) / planet_value)
    return threat_factor


def _rate_production(pp: float) -> float:
    return pp * RESOURCE_PRIORITY_MULTIPLIER * get_aistate().get_priority(PriorityType.RESOURCE_PRODUCTION)


def _rate_research(rp: float) -> float:
    # Research priority starts rather high, but in the long term it is usually much lower than production.
    return (
        rp
        * RESOURCE_PRIORITY_MULTIPLIER
        * min(
            get_aistate().get_priority(PriorityType.RESOURCE_PRODUCTION),
            get_aistate().get_priority(PriorityType.RESOURCE_RESEARCH),
        )
    )


def _rate_influence(ip: float) -> float:
    return ip * RESOURCE_PRIORITY_MULTIPLIER * get_aistate().get_priority(PriorityType.RESOURCE_INFLUENCE)


def _path_to_capital(system_id: SystemId):
    """Returns whether we know a path from our capital to that system. Always returns true if we have no capital."""
    universe = fo.getUniverse()
    capital_id = PlanetUtilsAI.get_capital()
    homeworld = universe.getPlanet(capital_id)
    if homeworld:
        home_system_id = homeworld.systemID
        least_jumps = universe.jumpDistance(home_system_id, system_id)
        return least_jumps != -1  # -1 means no path
    return True


def _planet_supply(planet: fo.planet) -> int:
    universe = fo.getUniverse()
    planet_supply = AIDependencies.supply_by_size.get(planet.size, 0)
    bld_types = {universe.getBuilding(bldg).buildingTypeName for bldg in planet.buildingIDs}.intersection(
        AIDependencies.building_supply
    )
    planet_supply += sum(
        AIDependencies.building_supply[bld_type].get(int(psize), 0)
        for psize in [-1, planet.size]
        for bld_type in bld_types
    )
    system = universe.getSystem(planet.systemID)
    supply_specials = set(planet.specials).union(system.specials).intersection(AIDependencies.SUPPLY_MOD_SPECIALS)
    planet_supply += sum(
        AIDependencies.SUPPLY_MOD_SPECIALS[_special].get(int(psize), 0)
        for _special in supply_specials
        for psize in [-1, planet.size]
    )
    return planet_supply


@cache_for_current_turn
# hashing entire planets? Perhaps better us a tuple of pids?
def _count_producers(candidates: tuple[fo.planet], min_stability: float, number: bool) -> float:
    """
    Return number of planets (number) or population on planets (not number) amongst candidates,
    which have industry focus and at least min_stability. Actually, since foci may change, planet currently
    having the focus are count 80% and planets having other foci are counted 30%.
    """
    value = 0.0
    for planet in candidates:
        if stability_with_focus(planet, FocusType.FOCUS_INDUSTRY) >= min_stability:
            focus_factor = 0.8 if planet.focus == FocusType.FOCUS_INDUSTRY else 0.3
            # TBD: perhaps use something between current and target population?
            # Surely not current, it will take time until we get the bonus we're evaluating...
            value += focus_factor * (1.0 if number else planet.currentMeterValue(fo.meterType.targetPopulation))
    return value


def _already_have_types(locally_owned_planets: list[fo.planet]) -> tuple[bool, bool]:
    """
    Determine whether list of planets contains asteroids and a gas giant.
    Gas giants are not considered, if they are populated by a species that dislikes GGG and the
    planet does not already have one built or enqueued.
    """
    have_asteroids = False
    have_gas_giant = False
    for planet in locally_owned_planets:
        if planet.type == fo.planetType.asteroids:
            have_asteroids = True
        if planet.type == fo.planetType.gasGiant:
            species = fo.getSpecies(planet.speciesName)
            ggg = BuildingType.GAS_GIANT_GEN
            # Currently, all gaseous beings dislike GGG, another one to place a GGG may be good
            if not species or ggg.value not in species.dislikes or planet.id in ggg.built_or_queued_at():
                have_gas_giant = True
    return have_asteroids, have_gas_giant


class _LocalPresence(NamedTuple):
    number: int  # number outposts + 3 x number of colonies
    asteroids_rating: float  # estimate how many colonies would profit from Microgravity Industry
    ggg_rating: float  # estimate how many colonies would profit from a gas giant generator


def _determine_local_presence(planet: fo.planet, details: list) -> _LocalPresence:
    universe = fo.getUniverse()
    planets = [universe.getPlanet(pid) for pid in get_owned_planets_in_system(planet.systemID) if pid != planet.id]
    populated = tuple(planet for planet in planets if planet.speciesName)
    details.append(f"local presence {len(planets)}({len(populated)} pop.)")
    asteroids_rating = ggg_rating = 0.0
    have_asteroids, have_gas_giant = _already_have_types(planets)
    if planet.type in (fo.planetType.gasGiant, fo.planetType.asteroids):
        if planet.type == fo.planetType.asteroids:
            if not have_asteroids and tech_is_complete(AIDependencies.PRO_MICROGRAV_MAN):
                users = _count_producers(populated, get_named_real("PRO_MICROGRAV_MAN_MIN_STABILITY"), True)
                bonus = get_named_real("PRO_MICROGRAV_MAN_TARGET_INDUSTRY_FLAT")
                asteroids_rating = _rate_production(users * bonus)
                details.append(f"Mirco-gravity users={users}, rating={asteroids_rating}")
        else:  # gas giant
            if not have_gas_giant and tech_is_complete(AIDependencies.PRO_ORBITAL_GEN):
                users = _count_producers(populated, get_named_real("BLD_GAS_GIANT_GEN_MIN_STABILITY"), True)
                bonus = get_named_real("BLD_GAS_GIANT_GEN_OUTPOST_TARGET_INDUSTRY_FLAT")
                ggg_rating = _rate_production(users * bonus)
                details.append(f"GGG users={users}, rating={ggg_rating}")
    return _LocalPresence(len(planets) + 2 * len(populated), asteroids_rating, ggg_rating)


def _check_solar_orbital_generator(system: fo.system, details: list) -> float:
    """
    Checks whether a planet in this system could allow us building a solar orbital generator.
    If possible, return a value for that possibility, otherwise return 0.
    """
    tech = AIDependencies.PRO_SOL_ORB_GEN
    rating = 0.0
    if system.starType in [fo.starType.blue, fo.starType.white]:
        factor = 0.0
        if tech_soon_available(tech, 5):
            bw_stars_claimed = count_claimed_stars(fo.starType.blue) + count_claimed_stars(fo.starType.white)
            if not bw_stars_claimed:
                factor = 1.0
                details.append(f"SOG(bright): {tech} and first bw star")
            else:
                # give a little bonus as backup
                factor = 1.0 / (bw_stars_claimed + 3)
                details.append(f"SOG(bright): {tech}, having {bw_stars_claimed} stars")
        if factor > 0.0:
            min_stability = get_named_real("BLD_SOL_ORB_GEN_MIN_STABILITY")
            gain_per_pop = get_named_real("BLD_SOL_ORB_GEN_BRIGHT_TARGET_INDUSTRY_PERPOP")
            users = _count_producers(get_empire_populated_planets(), min_stability, False)
            rating = _rate_production(factor * users * gain_per_pop)
            details.append(f"SOG(bright): users={users}, rating={rating}")
    return rating


def _check_blackhole_generator(system: fo.system, details: list) -> float:
    """
    Checks whether a planet in this system could allow us building a black hole generator.
    If possible, return a value for that possibility, otherwise return 0.
    """
    # Wait for Theory of Everything, because that is really expensive.
    # Once we have that, Singularity Generation shouldn't take too long, and we may as well prepare for it.
    tech1 = AIDependencies.LRN_EVERYTHING
    # red planets are only considered, if we could turn them into black holes
    tech2 = AIDependencies.LRN_ART_BLACK_HOLE
    tech3 = AIDependencies.PRO_SINGULAR_GEN
    rating = 0.0
    if tech_is_complete(tech1) and tech_soon_available(tech3, 5):
        factor = 0.0
        black_holes_claimed = count_claimed_stars(fo.starType.blackHole)
        if system.starType == fo.starType.blackHole:
            # blackholes with planets are pretty rare
            factor = 1.0 / (black_holes_claimed + 1)
            details.append(f"BHG: {tech1} and {black_holes_claimed} black holes yet")
        elif tech_is_complete(tech2):
            red_stars_claimed = count_claimed_stars(fo.starType.red)
            # we could turn this into a black hole
            factor = 1.0 / (2 * black_holes_claimed + red_stars_claimed + 1)
            details.append(f"BHG: {tech2}, {black_holes_claimed} black holes, {red_stars_claimed} reds")
        if not tech_is_complete(tech3):
            factor *= 0.5  # could not start building the BHG yet
        if factor:
            min_stability = get_named_real("BLD_BLACK_HOLE_POW_GEN_MIN_STABILITY")
            gain_per_pop = get_named_real("BLD_BLACK_HOLE_POW_GEN_TARGET_INDUSTRY_PERPOP")
            users = _count_producers(get_empire_populated_planets(), min_stability, False)
            rating = _rate_production(factor * gain_per_pop * users)
            details.append(f"BHG: users={users}, rating={rating}")
    return rating


def _check_neutronium_extraction(system: fo.system, details: list) -> float:
    """
    Checks whether a planet in this system could allow us building a neutronium extractor.
    If possible, return a value for that possibility, otherwise return 0.
    """
    rating = 0.0
    if system.starType == fo.starType.neutron:
        if tech_soon_available(AIDependencies.PRO_NEUTRONIUM_EXTRACTION, 5):
            neutron_stars_claimed = count_claimed_stars(fo.starType.neutron)
            # Note that a NEUTRONIUM_EXTRACTOR is much cheaper than a NEUTRONIUM_SYNTH, so even if can build
            # a synthesizer, getting a neutron star may help.
            built = len(BuildingType.NEUTRONIUM_SYNTH.built_at()) + len(BuildingType.NEUTRONIUM_EXTRACTOR.built_at())
            rating = 1.0 / (neutron_stars_claimed + built + 1)
        rating *= 5 * MINIMUM_COLONY_SCORE
        details.append(f"neutronium extraction={rating}")
    return rating


def _check_ruins(planet: fo.planet, details: list) -> float:
    """Check whether the planet has still unexplored ancients ruins and return a value for that."""
    rating = 0.0
    if AIDependencies.ANCIENT_RUINS_SPECIAL in planet.specials:
        rating = 1.0 if tech_is_complete(AIDependencies.LRN_XENOARCH) else 0.5
        # this can be very valuable!
        rating *= 10 * MINIMUM_COLONY_SCORE
        details.append(f"unexplored ancient ruins={rating}")
    return rating


def _check_asteroid_coating(planet: fo.planet, details: list) -> float:
    """Check whether the planet has an asteroid_coating special and return a value for that."""
    rating = 0.0
    # do we really need two different specials? Well, at the moment we have two...
    if (
        AIDependencies.ASTEROID_COATING_OWNED_SPECIAL in planet.specials
        or AIDependencies.ASTEROID_COATING_SPECIAL in planet.specials
    ):
        rating = 2 * MINIMUM_COLONY_SCORE
        if owned_asteroid_coatings():
            rating /= 3 * owned_asteroid_coatings()
        if not tech_is_complete(AIDependencies.SHP_ASTEROID_HULLS):
            rating /= 2
        details.append(f"asteroid coating={rating}")
    return rating


def _check_worldtree(planet: fo.planet, details: list) -> float:
    """Check whether the planet has an worldtree special and return a value for that."""
    rating = 0.0
    if not have_worldtree and AIDependencies.WORLDTREE_SPECIAL in planet.specials:
        rating = 0.5 * MINIMUM_COLONY_SCORE
        details.append(f"worldtree={rating}")
    return rating


def _check_nests(planet: fo.planet, details: list) -> float:
    """Check whether the planet has a nest special and return a value for that."""
    for special in planet.specials:
        if special_is_nest(special):
            nest_val = get_nest_rating(special)
            details.append(f"{special}={nest_val}")
            return nest_val
    return 0.0


def _rate_supply_extension(system: fo.system, planet_supply: int) -> float:
    """
    Determines how many supply groups an outpost with the given supply in the system would connect to.
    If this returns more than one, the outpost would potentially connect still disconnected planets.
    """
    if planet_supply < 1:
        return 0.0
    # planet_supply should be an int, better cast it just in case we get a float e.g. from a planet meter
    supplied = within_n_jumps(system.systemID, int(planet_supply))
    potentially_added = 0
    groups = set()
    for sys_id in supplied:
        if get_system_supply(sys_id) < 0:
            potentially_added += 1
        else:
            group = get_supply_group_id(sys_id)
            if not assertion_fails(group >= 0):
                groups.add(group)
    # supplying additional planets may be helpful, connecting planets to the capital helps a lot!
    groups_connected = max(len(groups) - 1, 0)
    capital_connected = 0
    if groups_connected:
        capital_group = get_supply_group_id(PlanetUtilsAI.get_capital_sys_id())
        if capital_group in groups:  # TODO: only count populated planets?
            capital_connected = sum(
                len(PlanetUtilsAI.get_planets_in__systems_ids(get_supply_group(g)))
                for g in groups
                if g != capital_group
            )
    if groups_connected:
        debug(
            f"Outpost at {system} with supply {planet_supply} connects {groups_connected} supply groups and "
            f"{capital_connected} planets to the capital"
        )
    return (0.05 * potentially_added + 0.2 * groups_connected + 0.5 * capital_connected) * MINIMUM_COLONY_SCORE


@cache_for_current_turn
def _rate_supply(system: fo.system, planet_supply: int) -> tuple[float, float]:
    """
    Evaluate supply situation. Return value for added supply and a factor to multiply the overall rating
    with that is lower for badly supplied planets.
    """
    sys_supply = get_system_supply(system.systemID)
    planet_supply = max(planet_supply, 0)  # there is no negative supply
    if sys_supply == 0:
        # get_system_supply == 0 means supplied, check if planet is on the border
        neighbors = get_neighbors(system.systemID)
        if all(supply_connected(sys_id, system.systemID) for sys_id in neighbors):
            factor = 1.0  # all neighbours are in the same supply group, so supply seems pretty save
        else:
            # planet is on the border of our supply network
            factor = 0.9 if planet_supply == 0 else 1.0
    else:
        # planet is outside the current supply network
        supply_sum = planet_supply + sys_supply
        if supply_sum >= 0:
            factor = min(1.0, 0.7 + 0.1 * supply_sum)  # >= 0.7
        else:
            factor = max(0.25, 0.6 + 0.1 * supply_sum)  # <= 0.5, planet would not be connected
    rating = _rate_supply_extension(system, planet_supply)
    return rating, factor


class _OutpostValues(NamedTuple):
    rating: float
    planet_supply: int
    supply_factor: float
    local_presence: int
    details: list


@cache_for_current_turn
def _calculate_outpost_rating(planet: fo.planet, supply_modifier: int) -> _OutpostValues:
    details = []
    empire = fo.getEmpire()
    universe = fo.getUniverse()
    system = universe.getSystem(planet.systemID)
    local_presence = _determine_local_presence(planet, details)
    planet_supply = _planet_supply(planet) + supply_modifier
    supply_rating, supply_factor = _rate_supply(system, planet_supply)
    if planet.owner == empire.empireID:
        # We already own it, so most outpost checks are pointless, but a species could extend supply range
        current_supply = int(planet.currentMeterValue(fo.meterType.maxSupply))
        current_supply_rating, _ = _rate_supply(system, current_supply)
        supply_rating -= current_supply_rating
        details.append(f"supply extension={planet_supply - current_supply}, rating={supply_rating}")
        return _OutpostValues(supply_rating, planet_supply, 1.0, local_presence.number, details)
    details.append(f"base_supply={planet_supply}, rating={supply_rating}, factor={supply_factor}")
    rating = (
        local_presence.asteroids_rating
        + local_presence.ggg_rating
        + local_presence.number * MINIMUM_COLONY_SCORE / 15
        + _check_solar_orbital_generator(system, details)
        + _check_blackhole_generator(system, details)
        + _check_neutronium_extraction(system, details)
        + _check_ruins(planet, details)  # TODO? AI could excavate them without supply connection
        + _check_asteroid_coating(planet, details)
        + _check_worldtree(planet, details)
        + _check_nests(planet, details)
        # TODO fortress?
    )
    rating = rating * supply_factor + supply_rating
    details.append(f"outpost_rating={rating}")
    return _OutpostValues(rating, planet_supply, supply_factor, local_presence.number, details)


def _check_cultural_buildings(planet: fo.planet, stability: float, max_population: float) -> tuple[float, float]:
    """
    Determines production and research would could get from cultural buildings.
    Checks for Cultural Archive, Cultural Library and Automatic History Analyser.
    """
    have_archive = have_library = have_analyser = False
    for building_id in planet.buildingIDs:
        have_archive |= BuildingType.CULTURE_ARCHIVES.is_this_type(building_id)
        have_library |= BuildingType.CULTURE_LIBRARY.is_this_type(building_id)
        have_analyser |= BuildingType.AUTO_HISTORY_ANALYSER.is_this_type(building_id)
    research = production = 0.0
    if have_archive and stability >= get_named_real("BLD_CULTURE_ARCHIVES_MIN_STABILITY"):
        research += get_named_real("BLD_CULTURE_ARCHIVES_TARGET_RESEARCH_FLAT")
        production += get_named_real("BLD_CULTURE_ARCHIVES_TARGET_INDUSTRY_PERPOP") * max_population
    # Most likely every planet having an archive also has a history analyser, but we could build one if it
    # only has the archive.
    if have_analyser or have_archive:
        research += get_named_real("BLD_AUTO_HISTORY_ANALYSER_TARGET_RESEARCH_FLAT")
    if have_library and stability >= get_named_real("BLD_CULTURE_LIBRARY_MIN_STABILITY"):
        research += get_named_real("BLD_CULTURE_LIBRARY_TARGET_RESEARCH_FLAT")
    return research, production


def _rate_focus_independent(planet: fo.planet, species: fo.species, stability: float, max_population: float) -> float:
    """
    Rate focus independent focuses production.
    """
    research = production = influence = 0.0
    if stability <= 0:
        return 0.0  # TBD: lower rating is barely above 0, since it would take very long to get the full bonuses?

    min_stability = get_named_real("PRO_ADAPTIVE_AUTO_MIN_STABILITY")
    if tech_soon_available(AIDependencies.PRO_AUTO_1, 3) and stability >= min_stability:
        production += get_named_real("PRO_ADAPTIVE_AUTO_TARGET_INDUSTRY_FLAT")

    # Nascent AI is such an early research, we will probably have it by the time the colony reached the stability
    if stability >= get_named_real("LRN_NASCENT_AI_MIN_STABILITY"):
        research += get_named_real("LRN_NASCENT_AI_TARGET_RESEARCH_FLAT")

    min_stability = get_named_real("ECCENTRIC_ORBIT_MIN_STABILITY")
    if "ECCENTRIC_ORBIT_SPECIAL" in planet.specials and stability >= min_stability:
        research += get_named_real("ECCENTRIC_ORBIT_TARGET_RESEARCH_FLAT")

    if (
        AIDependencies.Tags.ARTISTIC in species.tags
        and fo.getEmpire().policyAdopted("PLC_ARTISAN_WORKSHOPS")
        and stability >= get_named_real("ARTISANS_MIN_STABILITY_NO_FOCUS")
    ):
        # TBD? Strictly speaking, this is not focus independent, since it is not cumulative with the focused bonus
        influence += get_named_real("ARTISANS_INFLUENCE_FLAT_NO_FOCUS")

    cultural_research, cultural_production = _check_cultural_buildings(planet, stability, max_population)
    research += cultural_research
    production += cultural_production

    # TODO Philosopher Planet, goes both ways...

    rating = _rate_research(research) + _rate_production(production) + _rate_influence(influence)
    return rating


def _check_honeycomb(planet: fo.planet) -> float:
    # TBD more sophisticated calculation
    value = get_aistate().get_priority(PriorityType.RESOURCE_PRODUCTION) * 2
    if AIDependencies.HONEYCOMB_SPECIAL in planet.specials:
        if not have_honeycomb():
            return value
        else:
            return value / 3
    return 0


def _rate_with_industry_focus(
    planet: fo.planet,
    species: fo.species,
    max_pop: float,
    base_stability: float,
    details: list,
) -> float:
    focus = FocusType.FOCUS_INDUSTRY
    if focus not in species.foci:
        return 0.0
    stability = base_stability + PlanetUtilsAI.focus_stability_effect(species, focus)
    focus_rating = _rate_production(calculate_production(planet, species, max_pop, stability))
    flat_rating = _rate_focus_independent(planet, species, stability, max_pop)
    honey_rating = _check_honeycomb(planet)
    details.append(f"industry: stability={stability} focus={focus_rating}, flat={flat_rating}, honey={honey_rating}")
    return focus_rating + flat_rating + honey_rating


def _check_computronium(planet: fo.planet) -> float:
    # TBD more sophisticated calculation
    value = get_aistate().get_priority(PriorityType.RESOURCE_RESEARCH) * 1.5
    if AIDependencies.COMPUTRONIUM_SPECIAL in planet.specials:
        if not have_computronium():
            return value
        else:
            return value / 3
    return 0


def _rate_with_research_focus(
    planet: fo.planet,
    species: fo.species,
    max_pop: float,
    base_stability: float,
    details: list,
) -> float:
    focus = FocusType.FOCUS_RESEARCH
    if focus not in species.foci:
        return 0.0
    stability = base_stability + PlanetUtilsAI.focus_stability_effect(species, focus)
    focus_rating = _rate_research(calculate_research(planet, species, max_pop, stability))
    flat_rating = _rate_focus_independent(planet, species, stability, max_pop)
    computronium_rating = _check_computronium(planet)
    details.append(
        f"research: stability={stability} focus={focus_rating}, flat={flat_rating}, comp={computronium_rating}"
    )
    return focus_rating + flat_rating + computronium_rating


def _rate_with_influence_focus(
    planet: fo.planet,
    species: fo.species,
    max_pop: float,
    base_stability: float,
    details: list,
) -> float:
    focus = FocusType.FOCUS_INFLUENCE
    if focus not in species.foci:
        return 0.0
    stability = base_stability + PlanetUtilsAI.focus_stability_effect(species, focus)
    focus_rating = _rate_influence(calculate_influence(planet, species, max_pop, stability))
    flat_rating = _rate_focus_independent(planet, species, stability, max_pop)
    details.append(f"influence: stability={stability} focus={focus_rating}, flat={flat_rating}")
    return focus_rating + flat_rating


def _rate_growth_special(special: str) -> float:
    # habitable size affected by this special
    boost_value = empire_metabolisms.get(AIDependencies.metabolismBoosts[special], 0)
    multiplier = 1.0
    universe = fo.getUniverse()
    for pid in get_growth_specials().get(special, []):
        planet = universe.getPlanet(pid)
        if planet.focus == FocusType.FOCUS_GROWTH:
            # TODO check supply
            multiplier = 0.4  # we already have it, but a backup may still be useful
        else:
            species = fo.getSpecies(planet.speciesName)
            if species and FocusType.FOCUS_GROWTH in species.foci:
                multiplier = min(multiplier, 0.9)  # planet is currently not exporting it, but it could
    return boost_value * multiplier * MINIMUM_COLONY_SCORE * 0.3


def _rate_with_growth_focus(
    planet: fo.planet,
    species: fo.species,
    max_pop: float,
    base_stability: float,
    details: list,
) -> float:
    focus = FocusType.FOCUS_GROWTH
    if focus not in species.foci:
        return 0.0
    stability = base_stability + PlanetUtilsAI.focus_stability_effect(species, focus)
    focus_rating = 0.0
    for special in [spec for spec in planet.specials if spec in AIDependencies.metabolismBoosts]:
        rating = _rate_growth_special(special)
        focus_rating += rating
        details.append(f"Bonus for {special}: {rating:.1f}")
    flat_rating = _rate_focus_independent(planet, species, stability, max_pop)
    details.append(f"growth: stability={stability} focus={focus_rating}, flat={flat_rating}")
    return focus_rating + flat_rating


def _rate_planet_vision(species: fo.species, supply_factor: float, details: list) -> float:
    """
    Give a small bonus to species with better detection.
    """
    detection_val = detection_value(species.name)  # -1 to 3
    # Well supplied planets are likely surrounded by other colonies. Vision is more important in the outskirts.
    # Could do some more sophisticated calculations here, but there or more important factors...
    supply_val = 1.05 - max(supply_factor, 0.8)  # 0.5 to 0.25
    scaling = MINIMUM_COLONY_SCORE * 0.3
    result = detection_val * supply_val * scaling
    if result:
        details.append(f"detection bonus:{result:.1f}")
    return result


def _adapt_population_for_rating(planet: fo.planet, max_population, details: list) -> float:
    current_population = planet.currentMeterValue(fo.meterType.population)
    if not current_population:
        current_population = 3.0 if tech_is_complete("GRO_LIFECYCLE_MAN") else 1.0
    # If more than half full, we consider its full potential population
    # TODO: could do some calculations to predict size in e.g. 20 turns
    result = min((2 * current_population + max_population) / 2, max_population)
    details.append(f"Adapted population: {result}")
    return result


def _calculate_colony_rating(planet: fo.planet, species: fo.species, supply_factor: float, details: list) -> float:
    dummy = []
    max_pop = calc_max_pop(planet, species, dummy)  # too much detail...
    if max_pop <= 0:
        details.append(f"{species.name} wouldn't survive")
        return 0.0
    details.append(f"max_pop={max_pop}")

    max_pop = _adapt_population_for_rating(planet, max_pop, details)
    base_stability = calculate_stability(planet, species)
    industry_rating = _rate_with_industry_focus(planet, species, max_pop, base_stability, details)
    research_rating = _rate_with_research_focus(planet, species, max_pop, base_stability, details)
    influence_rating = _rate_with_influence_focus(planet, species, max_pop, base_stability, details)
    growth_rating = _rate_with_growth_focus(planet, species, max_pop, base_stability, details)
    options = sorted((industry_rating, research_rating, influence_rating, growth_rating))
    best = options.pop()
    second = options.pop()
    rating = best + max(0.0, second - best * 0.5)  # if second best is more than half of best, increase rating
    pilot_rating = rate_colony_for_pilots(planet, species, details)
    vision_rating = _rate_planet_vision(species, supply_factor, details)
    rating += pilot_rating + vision_rating
    # TODO: Xenophobic harassment (by this planet and by others)
    details.append(
        f"industry: {industry_rating}, research: {research_rating}, influence: {influence_rating}, "
        f"growth: {growth_rating}, pilots: {pilot_rating} => rating={rating}"
    )
    return rating


def _adapt_supply_factor(supply_factor: float, species: fo.species) -> float:
    """
    Limit the supply factor (penalty for bad supply) for INDEPENDENT species, especially Sly, which have basically
    no other chance then to settle in disconnected places for a long time.
    """
    if AIDependencies.Tags.INDEPENDENT in species.tags:
        current_turn = fo.currentTurn()
        if species.name == "SP_SLY":
            return max(supply_factor, 1.0 - current_turn * 0.003)  # minimum 0.4 at turn 200
        else:
            return max(supply_factor, 0.9 - current_turn * 0.005)  # minimum 0.4 at turn 100
    else:
        return supply_factor


def _colony_upkeep(num_exobots: int, num_normal: int, num_outpost: int) -> float:
    colonies = num_exobots + num_normal - 1  # assume we have a capital, calculation won't be too wrong otherwise
    per_colony = get_named_real("COLONY_ADMIN_COSTS_PER_PLANET")
    outpost_factor = get_named_real("OUTPOST_RELATIVE_ADMIN_COUNT")
    # see influence.macros, exobot colonies pay like a colony, but count as outposts for the cost increase
    return colonies * per_colony * (num_normal + outpost_factor * (num_exobots + num_outpost)) ** 0.5


def _rate_upkeep(planet: fo.planet, species_name: SpeciesName, details: list) -> float:
    exobot = SpeciesName("SP_EXOBOT")
    num_exobots = len(get_empire_planets_by_species().get(exobot, []))
    num_normal = len(get_empire_populated_planets()) - num_exobots
    num_outpost = len(get_empire_outposts())
    current_upkeep = _colony_upkeep(num_exobots, num_normal, num_outpost)
    if species_name == exobot:
        num_exobots += 1
    elif not species_name:
        num_outpost += 1
    else:
        num_normal += 1
    ip_cost = current_upkeep - _colony_upkeep(num_exobots, num_normal, num_outpost)
    species = fo.getSpecies(species_name)
    if species and planet.id in species.homeworlds:
        ip_cost -= AIDependencies.HOMEWORLD_INFLUENCE_COST
    result = _rate_influence(ip_cost)
    details.append(f"IP upkeep: {ip_cost} -> {result}")
    return result


def _calculate_planet_rating(planet: fo.planet, species_name: SpeciesName) -> tuple[float, list]:
    if not planet:
        return 0.0, ["unknown planet"]
    if not _path_to_capital(planet.systemID):
        return 0.0, ["no path from capital"]
    species = None
    if species_name:
        species = fo.getSpecies(species_name)
        if assertion_fails(species, f"unknown species {species_name}"):
            return 0.0, [f"unknown species {species_name}"]
        supply_modifier = get_species_supply(species_name)
    else:
        supply_modifier = 0
    sys_partial_vis_turn = get_partial_visibility_turn(planet.systemID)
    planet_partial_vis_turn = get_partial_visibility_turn(planet.id)
    # TODO: check whether this copied code makes sense
    if planet_partial_vis_turn < sys_partial_vis_turn:
        # last time we had partial vis of the system, the planet was stealthed to us
        return 0.0, ["Couldn't get current info, planet was stealthed at last sighting"]

    rating, planet_supply, supply_factor, local_presence, details = _calculate_outpost_rating(planet, supply_modifier)
    if species:
        colony_rating = _calculate_colony_rating(planet, species, supply_factor, details)
        supply_factor = _adapt_supply_factor(supply_factor, species)
        if colony_rating <= 0.0:
            return 0.0, details  # planet not habitable for species? skip the rest
        rating += colony_rating * supply_factor
    threat_factor = _determine_colony_threat_factor(planet, species_name, local_presence)
    if threat_factor < 1.0:
        threat_factor = _revise_threat_factor(threat_factor, rating, planet.systemID)
        rating *= threat_factor
        details.append("threat reducing value by %3d %%" % (100 * (1 - threat_factor)))
    rating += _rate_upkeep(planet, species_name, details)
    return rating, details



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_PHYS_BRAIN",
    description="LRN_PHYS_BRAIN_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=10 * TECH_COST_MULTIPLIER,
    researchturns=2,
    tags=["PEDIA_LEARNING_CATEGORY"],
    unlock=[Item(type=UnlockPolicy, name="PLC_LIBERTY"), Item(type=UnlockBuilding, name="BLD_AUTO_HISTORY_ANALYSER")],
    graphic="icons/tech/the_physical_brain.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_ORGANIC_WAR_ADAPTION",
    description="SHP_ORGANIC_WAR_ADAPTION_DESC",
    short_description="SHP_ORGANIC_WAR_ADAPTION_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=120 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_SR_WEAPON_TECHS"],
    prerequisites=["SHP_WEAPON_2_2", "SHP_ORG_HULL"],
    unlock=Item(type=UnlockShipPart, name="SP_SOLAR_CONCENTRATOR"),
    graphic="icons/ship_parts/solarcollector.png",
)



from focs._effects import (
    EffectsGroup,
    Happiness,
    NamedReal,
    OwnedBy,
    Planet,
    SetTargetResearch,
    Source,
    TargetPopulation,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.priorities import TARGET_AFTER_2ND_SCALING_PRIORITY

Tech(
    name="LRN_NASCENT_AI",
    description="LRN_NASCENT_AI_DESC",
    short_description="RESEARCH_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=48 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_LEARNING_CATEGORY"],
    prerequisites=["LRN_ALGO_ELEGANCE"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & TargetPopulation(low=0.0001)
            & Happiness(low=NamedReal(name="LRN_NASCENT_AI_MIN_STABILITY", value=10)),
            priority=TARGET_AFTER_2ND_SCALING_PRIORITY,
            effects=SetTargetResearch(value=Value + NamedReal(name="LRN_NASCENT_AI_TARGET_RESEARCH_FLAT", value=1)),
        )
    ],
    graphic="icons/tech/basic_autolabs.png",
)



from focs._effects import (
    CurrentContent,
    CurrentTurn,
    EffectsGroup,
    GenerateSitRepMessage,
    IsSource,
    LocalCandidate,
    Max,
    MaxSecondaryStat,
    NamedReal,
    OwnedBy,
    PartCapacity,
    PartSecondaryStat,
    SetCapacity,
    SetMaxCapacity,
    SetMaxSecondaryStat,
    SetSecondaryStat,
    Ship,
    ShipPartMeter,
    Source,
    Statistic,
    Target,
    TurnTechResearched,
    Value,
)
from macros.misc import (
    FIGHTER_DAMAGE_FACTOR,
    SHIP_WEAPON_DAMAGE_FACTOR,
)
from macros.priorities import AFTER_ALL_TARGET_MAX_METERS_PRIORITY
from techs.techs import (
    ARBITRARY_BIG_NUMBER_FOR_METER_TOPUP,
    EMPIRE_OWNED_SHIP_WITH_PART,
    SHIP_PART_UPGRADE_RESUPPLY_CHECK,
)

# Tech based setting of damage and number of shots of a weapon. Registers named reals for use in the pedia.
# Be careful the effect does not show in the ship damage/shot estimates in the ship designer UI if the tech is not researched yet.
# @1@ part name


def WEAPON_BASE_EFFECTS(part_name: str):
    # these NamedReals need to be parsed to be registered in the pedia. XXX remove this when the pedia can look up ship meters/part capacity
    NamedReal(name=f"{part_name}_PART_CAPACITY", value=PartCapacity(name=part_name))
    NamedReal(name=f"{part_name}_PART_SECONDARY_STAT", value=PartSecondaryStat(name=part_name))

    # Set the damage and shot meters to the max values every turn.
    # The topup_effects_group is necessary..
    # 1) the first turn after an upgrade is researched, since the resupply currently takes place in a portion of the turn before meters are updated
    # 2) for any arbitrary "temporary" change of Max values (e.g. PLC_FLANKING); "permanent" changes could be modeled differently
    topup_effects_group = EffectsGroup(
        scope=EMPIRE_OWNED_SHIP_WITH_PART(part_name),
        accountinglabel=part_name,
        priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=[
            SetCapacity(partname=part_name, value=Value + ARBITRARY_BIG_NUMBER_FOR_METER_TOPUP),
            SetSecondaryStat(partname=part_name, value=Value + ARBITRARY_BIG_NUMBER_FOR_METER_TOPUP),
        ],
    )
    return [topup_effects_group]


# Tech Upgrade a direct weapon damage each turn if it was resupplied after the tech was researched.
# Be careful this does show in the ship damage/shot estimates in the ship designer UI if the tech is not researched.
# This can correctly estimate the total sum of all related weapon techs up to this tech if
# * following the tech naming scheme TECH_NAME_X with X being the level of the upgrade (and X being in 2..9)
# * and all tech level upgrade give the same amount of upgrade ( so the total sum is X * unscaled_upgrade)
# @1@ tech name
# @2@ part name
# @3@ unscaled_upgrade gets added to max capacity after scaling it with SHIP_WEAPON_DAMAGE_FACTOR
# @4@ upgraded_damage_override if given overrides the total sum over the researched techs shown in the upgrade info sitrep
def WEAPON_UPGRADE_CAPACITY_EFFECTS(
    tech_name: str, part_name: str, unscaled_upgrade: int, upgraded_damage_override: int = -1
):
    # the following recursive lookup works, but is not acceptable because of delays. as long as the parser is sequential, the parallel waiting feature is kind of a bug
    # previous_upgrade_effect = PartCapacity(name=part_name) if (tech_name[-1] == "2") else NamedRealLookup(name = tech_name[0:-1] + "2_UPGRADED_DAMAGE")  # + str(int(tech_name[-1]) - 1))
    upgraded_damage = (
        upgraded_damage_override if upgraded_damage_override != -1 else unscaled_upgrade * (int(tech_name[-1]) - 1)
    )
    return [
        EffectsGroup(
            scope=EMPIRE_OWNED_SHIP_WITH_PART(part_name) & SHIP_PART_UPGRADE_RESUPPLY_CHECK(CurrentContent),
            accountinglabel=part_name,
            effects=SetMaxCapacity(partname=part_name, value=Value + unscaled_upgrade * SHIP_WEAPON_DAMAGE_FACTOR),
        ),
        # Inform the researching empire that ships in supply will get upgraded before next combat
        EffectsGroup(
            scope=IsSource,
            activation=(CurrentTurn == TurnTechResearched(empire=Source.Owner, name=CurrentContent)),
            effects=GenerateSitRepMessage(
                message="SITREP_WEAPONS_UPGRADED",
                label="SITREP_WEAPONS_UPGRADED_LABEL",
                icon="icons/sitrep/weapon_upgrade.png",
                parameters={
                    "empire": Source.Owner,
                    "shippart": part_name,
                    "tech": CurrentContent,
                    # str(CurrentContent) -> <ValueRefString object at 0x...>
                    "dam": NamedReal(
                        name=tech_name + "_UPGRADE_DAMAGE", value=unscaled_upgrade * SHIP_WEAPON_DAMAGE_FACTOR
                    ),
                    "sum": NamedReal(
                        name=tech_name + "_UPGRADED_DAMAGE",
                        value=PartCapacity(name=part_name) + (SHIP_WEAPON_DAMAGE_FACTOR * upgraded_damage),
                    ),
                },
                empire=Target.Owner,
            ),
        ),
    ]


# Tech Upgrade a direct weapon number of shots each turn if it was resupplied after the tech was researched.
# Be careful this does show in the ship damage/shot estimates in the ship designer UI if the tech is not researched.
# @1@ tech name
# @2@ part name
# @3@ extra_shots gets added to max secondary stat
def WEAPON_UPGRADE_SECONDARY_STAT_EFFECTS(tech_name: str, part_name: str, extra_shots: int):
    return [
        EffectsGroup(
            scope=EMPIRE_OWNED_SHIP_WITH_PART(part_name) & SHIP_PART_UPGRADE_RESUPPLY_CHECK(CurrentContent),
            accountinglabel=part_name,
            effects=SetMaxSecondaryStat(partname=part_name, value=Value + extra_shots),
        ),
        # Inform the researching empire that ships in supply will get upgraded before next combat
        # TODO use conditional If effect to print a different message using a different shotsum e.g. "?" when there are no ships built yet
        #      or do value_ref<string> ops (which does not exist yet AFAIK)
        #         ("?" * StatisticIf(float, ~( Ship & OwnedBy(empire = Source.Owner) & DesignHasPart(part_name) ) )) + Statistic(...
        EffectsGroup(
            scope=IsSource,
            activation=(CurrentTurn == TurnTechResearched(empire=Source.Owner, name=CurrentContent)),
            effects=GenerateSitRepMessage(
                message="SITREP_WEAPON_SHOTS_UPGRADED",
                label="SITREP_WEAPON_SHOTS_UPGRADED_LABEL",
                icon="icons/sitrep/weapon_upgrade.png",
                parameters={
                    "empire": Source.Owner,
                    "shippart": part_name,
                    "tech": CurrentContent,
                    "shots": NamedReal(
                        name=tech_name + "_UPGRADE_SHOTS",
                        value=extra_shots,
                    ),
                    # a part value is not changeable. In the backend we also do not know which techs apply to a part
                    # options. Also note we cant register this as a named value ref as the value changes.
                    # 0) count the bonus in an empire meter
                    # 1) register all relevant tech names before calling this; and check here if those are researched
                    #    a named value ref to a list of strings could help for lazy eval
                    # 2) trying to get the new value from the highest known one - note this only works if a ship with the part is built
                    # 3) forget about it
                    "shotsum": extra_shots
                    + Statistic(
                        float,
                        Max,
                        value=ShipPartMeter(part=part_name, meter=MaxSecondaryStat, ship=LocalCandidate.ID),
                        condition=Ship & OwnedBy(empire=Source.Owner),
                    ),
                },
                empire=Target.Owner,
            ),
        ),
    ]


# Tech Upgrade SetMaxSecondaryStat effect for hangar part damage
# @1@ tech name
# @2@ part name
# @3@ base_damage_bonus scaled by figher damage gets added to max secondary stat
def HANGAR_UPGRADE_SECONDARY_STAT_EFFECT(tech_name: str, part_name: str, base_damage_bonus: int):
    return SetMaxSecondaryStat(
        partname=part_name,
        value=Value
        + NamedReal(
            name=tech_name + "_" + part_name + "_DAMAGE_BONUS", value=base_damage_bonus * FIGHTER_DAMAGE_FACTOR
        ),
    )



from focs._effects import EffectsGroup, Huge, IsSource, Large, Medium, Planet, SetPlanetSize, Small, Tiny, Turn

LARGE_PLANET = [
    EffectsGroup(scope=IsSource, activation=Planet() & Turn(high=0), effects=SetPlanetSize(planetsize=Large)),
]

NOT_LARGE_PLANET = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Turn(high=0) & ~Planet(size=[Tiny, Medium]),
        effects=SetPlanetSize(planetsize=Small),
    )
]

HUGE_PLANET = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Turn(high=0),
        effects=SetPlanetSize(planetsize=Huge),
    )
]



from focs._effects import (
    EffectsGroup,
    Focus,
    GalaxyMaxAIAggression,
    Happiness,
    IsHuman,
    IsSource,
    Planet,
    SetTargetResearch,
    Target,
    Value,
)
from macros.base_prod import RESEARCH_PER_POP
from macros.misc import DESCRIPTION_EFFECTSGROUP_MACRO
from macros.multiplier import (
    BAD_MULTIPLIER,
    GOOD_MULTIPLIER,
    GREAT_MULTIPLIER,
    ULTIMATE_MULTIPLIER,
    VERY_BAD_MULTIPLIER,
)
from macros.priorities import (
    TARGET_AFTER_SCALING_PRIORITY,
    TARGET_EARLY_BEFORE_SCALING_PRIORITY,
    TARGET_SCALING_PRIORITY,
)

_BASIC_RESEARCH = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_RESEARCH"]) & Happiness(low=0),
        accountinglabel="FOCUS_RESEARCH_LABEL",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetTargetResearch(value=Value + Target.Population * RESEARCH_PER_POP),
    ),
    EffectsGroup(  # gives human bonuses when AI Aggression set to Beginner
        scope=IsSource,
        activation=Planet() & IsHuman & (GalaxyMaxAIAggression == 0),  # human player, not human species
        accountinglabel="DIFFICULTY",
        priority=TARGET_AFTER_SCALING_PRIORITY,
        effects=SetTargetResearch(value=Value * 2),
    ),
]

NO_RESEARCH = DESCRIPTION_EFFECTSGROUP_MACRO("NO_RESEARCH_DESC")


VERY_BAD_RESEARCH = [
    *_BASIC_RESEARCH,
    EffectsGroup(
        description="VERY_BAD_RESEARCH_DESC",
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_RESEARCH"]),
        accountinglabel="VERY_BAD_RESEARCH_LABEL",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetTargetResearch(value=Value * VERY_BAD_MULTIPLIER),
    ),
]


BAD_RESEARCH = [
    *_BASIC_RESEARCH,
    EffectsGroup(
        description="BAD_RESEARCH_DESC",
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_RESEARCH"]),
        accountinglabel="BAD_RESEARCH_LABEL",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetTargetResearch(value=Value * BAD_MULTIPLIER),
    ),
]

AVERAGE_RESEARCH = _BASIC_RESEARCH

GOOD_RESEARCH = [
    *_BASIC_RESEARCH,
    EffectsGroup(
        description="GOOD_RESEARCH_DESC",
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_RESEARCH"]) & Happiness(low=2),
        accountinglabel="GOOD_RESEARCH_LABEL",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetTargetResearch(value=Value * GOOD_MULTIPLIER),
    ),
]


GREAT_RESEARCH = [
    *_BASIC_RESEARCH,
    EffectsGroup(
        description="GREAT_RESEARCH_DESC",
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_RESEARCH"]) & Happiness(low=0),
        accountinglabel="GREAT_RESEARCH_LABEL",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetTargetResearch(value=Value * GREAT_MULTIPLIER),
    ),
]


ULTIMATE_RESEARCH = [
    *_BASIC_RESEARCH,
    EffectsGroup(
        description="ULTIMATE_RESEARCH_DESC",
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_RESEARCH"]),
        accountinglabel="ULTIMATE_RESEARCH_LABEL",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetTargetResearch(value=Value * ULTIMATE_MULTIPLIER),
    ),
]



from logging import error

normalization_dict = {
    "empire": "empire_object",
    "int": "number",
    "str": "string",
    "float": "floating_number",
    "object": "obj",
    "IntBoolMap": "int_bool_map",
    "IntDblMap": "int_dbl_map",
    "universeObject": "base_object",
    "meterType": "meter_type",
    "bool": "boolean",
    "StringVec": "string_list",
    "shipDesign": "ship_design",
    "universe": "universe_object",
    "researchQueue": "research_queue",
    "resPoolMap": "res_pool",
    "productionQueue": "production_queue",
    "diplomaticMessage": "diplomatic_message",
    "ship": "ship_object",
    "species": "species_object",
    "planetType": "planet_type",
    "system": "system_object",
    "tech": "tech_object",
    "list": "item_list",
    "planet": "planet_object",
    "shipPart": "ship_part",
    "resPool": "res_pool",
    "researchQueueElement": "research_queue_element",
    "shipSlotType": "ship_slot_type",
    "IntIntMap": "int_int_map",
    "IntPairVec": "int_pair_list",
    "IntSet": "int_set",
    "IntSetSet": "int_set_set",
    "IntVec": "int_list",
    "IntVisibilityMap": "int_visibility_map",
    "UnlockableItemVec": "unlockable_item_vec",
    "StringSet": "string_set",
    "VisibilityIntMap": "visibility_int_map",
    "buildingType": "buildingType",
    "productionQueueElement": "production_queue_element",
    "resourceType": "resource_type",
    "fleetAggression": "fleet_aggression",
    "buildType": "build_type",
    "field": "field",
    "shipHull": "ship_hull",
    "PlanetSize": "planet_size",
    "planetSize": "planet_size",
    "unlockableItemType": "unlockable_item_type",
    "dict": "dictionary",
    "StarType": "star_type",
    "starType": "star_type",
    "UnlockableItemType": "unlocable_item_type",
    "FleetPlan": "fleet_plan",
    "MeterTypeMeterMap": "meter_type_meter_map",
    "PairIntInt_IntMap": "pair_int_int_int_map",
    "MonsterFleetPlan": "monster_fleet_plan",
    "ShipPartMeterMap": "ship_part_meter_map",
    "ShipSlotVec": "ship_slot_vec",
    "special": "special",
    "IntFltMap": "int_flt_map",
    "ruleType": "rule_type",
    "influenceQueueElement": "influence_queue_element",
}


def report_once_and_register(argument_type):
    error("Can't find proper name for: %s, please add it to name mapping \n" % argument_type)
    normalization_dict[argument_type] = "arg"


def normalize_name(tp):
    argument_type, provided_name = tp
    if not provided_name.startswith("arg"):
        return provided_name
    if argument_type not in normalization_dict:
        report_once_and_register(argument_type)
    return normalization_dict[argument_type]



from focs._effects import (
    EffectsGroup,
    Focus,
    Happiness,
    NamedReal,
    OwnedBy,
    Planet,
    SetTargetIndustry,
    Source,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import INDUSTRY_PER_POP, TECH_COST_MULTIPLIER
from macros.priorities import TARGET_AFTER_SCALING_PRIORITY

Tech(
    name="PRO_ROBOTIC_PROD",
    description="PRO_ROBOTIC_PROD_DESC",
    short_description="INDUSTRY_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=24 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & Focus(type=["FOCUS_INDUSTRY"])
            & Happiness(low=NamedReal(name="PRO_ROBOTIC_PROD_MIN_STABILITY", value=5)),
            priority=TARGET_AFTER_SCALING_PRIORITY,
            effects=SetTargetIndustry(
                value=Value
                + Target.Population
                * NamedReal(name="PRO_ROBOTIC_PROD_TARGET_INDUSTRY_PERPOP", value=0.25 * INDUSTRY_PER_POP)
            ),
        )
    ],
    graphic="icons/tech/robotic_production.png",
)



from collections.abc import Mapping

from common.fo_typing import SystemId, Turn
from stub_generator.stub_generator.collection_classes import make_type
from stub_generator.stub_generator.rtype.mapper import Mapper
from stub_generator.stub_generator.rtype.utils import get_name_for_mapping

_method_map = Mapper(
    "methods",
    {
        "supplyProjections": Mapping[SystemId, int],
        "turnPolicyAdopted": Turn,
        "name": str,
    },
)


def update_method_rtype(method_name: str, rtype: str) -> str:
    if method_name in _method_map:
        return get_name_for_mapping(_method_map[method_name])
    else:
        return make_type(rtype)



from functools import wraps
from typing import Callable


class InitializerLock:
    """
    Class that ensure that it's properly initialized before the first use of data.
    """

    def __init__(self, name):
        self.__initialized = False
        self.__name = name

    def lock(self):
        self.__initialized = False

    def unlock(self):
        self.__initialized = True

    def __call__(self, function: Callable):
        @wraps(function)
        def wrapper(*args, **kwargs):
            if self.__initialized:
                return function(*args, **kwargs)
            else:
                raise ValueError(
                    f"Call of '{function.__name__}' is forbidden before {self.__class__.__name__}('{self.__name}') is initialized"
                )

        return wrapper



from focs._effects import (
    EffectsGroup,
    Focus,
    GalaxyMaxAIAggression,
    Happiness,
    IsHuman,
    IsSource,
    NamedReal,
    Planet,
    SetTargetIndustry,
    Target,
    TargetIndustry,
    Value,
)
from macros.base_prod import INDUSTRY_PER_POP
from macros.priorities import (
    TARGET_AFTER_SCALING_PRIORITY,
    TARGET_EARLY_BEFORE_SCALING_PRIORITY,
    TARGET_SCALING_PRIORITY,
)

_BASIC_INDUSTRY = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & TargetIndustry(low=0) & Happiness(low=0) & Focus(type=["FOCUS_INDUSTRY"]),
        accountinglabel="FOCUS_INDUSTRY_LABEL",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetTargetIndustry(
            value=Value
            + Target.Population * NamedReal(name="INDUSTRY_FOCUS_TARGET_INDUSTRY_PERPOP", value=1.0 * INDUSTRY_PER_POP)
        ),
    ),
    EffectsGroup(  # gives human bonuses when AI Aggression set to Beginner
        scope=IsSource,
        activation=Planet() & IsHuman & (GalaxyMaxAIAggression == 0),  # human player, not human species
        accountinglabel="DIFFICULTY",
        priority=TARGET_AFTER_SCALING_PRIORITY,
        effects=SetTargetIndustry(value=Value * 2),
    ),
]


def _industry(tag, default_multiplier):
    return [
        *_BASIC_INDUSTRY,
        EffectsGroup(
            description=f"{tag}_INDUSTRY_DESC",
            scope=IsSource,
            activation=Planet() & TargetIndustry(low=0) & Happiness(low=0) & Focus(type=["FOCUS_INDUSTRY"]),
            accountinglabel=f"{tag}_INDUSTRY_LABEL",
            priority=TARGET_SCALING_PRIORITY,
            effects=SetTargetIndustry(
                value=Value * NamedReal(name=f"{tag}_INDUSTRY_TARGET_INDUSTRY_SCALING", value=default_multiplier)
            ),
        ),
    ]



from focs._effects import (
    Conditional,
    EffectsGroup,
    LocalCandidate,
    MaxOf,
    MinOf,
    NamedReal,
    NamedRealLookup,
    OwnedBy,
    Planet,
    SetConstruction,
    SetIndustry,
    SetResearch,
    SetStockpile,
    Source,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.priorities import AFTER_ALL_TARGET_MAX_METERS_PRIORITY

Tech(
    name="CON_FRC_ENRG_STRC",
    description="CON_FRC_ENRG_STRC_DESC",
    short_description="METER_GROWTH_SHORT_DESC",
    category="CONSTRUCTION_CATEGORY",
    researchcost=350 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_CONSTRUCTION_CATEGORY"],
    prerequisites=["LRN_FORCE_FIELD", "CON_ARCH_PSYCH"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            accountinglabel="CON_TECH_ACCOUNTING_LABEL",
            priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
            effects=[
                Conditional(
                    condition=(Value(LocalCandidate.Industry) <= Value(LocalCandidate.TargetIndustry)),
                    effects=[
                        SetIndustry(
                            value=MinOf(
                                float,
                                Value + NamedReal(name="FORCE_ENERGY_RATE_INCREASE", value=3.0),
                                Value(Target.TargetIndustry),
                            )
                        )
                    ],
                    else_=[
                        SetIndustry(
                            value=MaxOf(
                                float,
                                Value - NamedRealLookup(name="FORCE_ENERGY_RATE_INCREASE"),
                                Value(Target.TargetIndustry),
                            )
                        )
                    ],
                ),
                Conditional(
                    condition=(Value(LocalCandidate.Research) <= Value(LocalCandidate.TargetResearch)),
                    effects=[
                        SetResearch(
                            value=MinOf(
                                float,
                                Value + NamedRealLookup(name="FORCE_ENERGY_RATE_INCREASE"),
                                Value(Target.TargetResearch),
                            )
                        )
                    ],
                    else_=[
                        SetResearch(
                            value=MaxOf(
                                float,
                                Value - NamedRealLookup(name="FORCE_ENERGY_RATE_INCREASE"),
                                Value(Target.TargetResearch),
                            )
                        )
                    ],
                ),
                Conditional(
                    condition=(Value(LocalCandidate.Construction) <= Value(LocalCandidate.TargetConstruction)),
                    effects=[
                        SetConstruction(
                            value=MinOf(
                                float,
                                Value + NamedRealLookup(name="FORCE_ENERGY_RATE_INCREASE"),
                                Value(Target.TargetConstruction),
                            )
                        )
                    ],
                    else_=[
                        SetConstruction(
                            value=MaxOf(
                                float,
                                Value - NamedRealLookup(name="FORCE_ENERGY_RATE_INCREASE"),
                                Value(Target.TargetConstruction),
                            )
                        )
                    ],
                ),
                Conditional(
                    condition=(Value(LocalCandidate.Stockpile) <= Value(LocalCandidate.MaxStockpile)),
                    effects=[
                        SetStockpile(
                            value=MinOf(
                                float,
                                Value + NamedRealLookup(name="FORCE_ENERGY_RATE_INCREASE"),
                                Value(Target.MaxStockpile),
                            )
                        )
                    ],
                    else_=[
                        SetStockpile(
                            value=MaxOf(
                                float,
                                Value - NamedRealLookup(name="FORCE_ENERGY_RATE_INCREASE"),
                                Value(Target.MaxStockpile),
                            )
                        )
                    ],
                ),
            ],
        )
    ],
    graphic="icons/tech/force_energy_structures.png",
)



import os
import sys

current_dir = os.path.dirname(__file__)
common = os.path.join(current_dir, "..", "..", "..")
assert os.path.exists(common)
sys.path.append(common)

charting = os.path.join(current_dir, "..", "..")
assert os.path.exists(charting)
sys.path.append(charting)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_SPATIAL_DISTORT_GEN",
    description="LRN_SPATIAL_DISTORT_GEN_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=200 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_LEARNING_CATEGORY"],
    prerequisites=["LRN_NDIM_SUBSPACE"],
    unlock=Item(type=UnlockBuilding, name="BLD_SPATIAL_DISTORT_GEN"),
    graphic="icons/tech/controlled_gravity_wells.png",
)



"""This package provides json-based encoding and decoding for use in FreeOrion AI savegames.

The encoding is json-based with custom prefixes to support some objects
and dictionary keys of types which are not supported in standard json.

The decoder will only load trusted classes as defined in _definitions.py,
if an unknown/untrusted object is encountered, it will raise a InvalidSaveGameException.

When class instances are loaded, the __setstate__ method will be invoked if available or
the __dict__ content will be set directly. It is the responsibility of each trusted class
to define a __setstate__ method to verify and possibly sanitize the passed data.
The __setstate__ method is expected to raise an Exception if invalid data was passed.

In contrast to standard library json module, the type of dict keys (e.g. int) is preserved:
    import json
    json_string = json.dumps({1: 2})  ==> '{"1": 2}'
    json.loads(json_string) ==> {"1": 2}

    import savegame_codec  # load package
    json_string = savegame_codec.encode({1: 2})  ==> '{"__INT__1": 2'}'
    savegame_codec.decode(json_string)  ==> {1: 2}


Public functions:
    encode  -  encode a python object to an unambigious string representation
    decode  -  decode and interpret a json string to retreive python objects
    build_savegame_string  - encode the AIstate and return the zlib-compressed result
    load_savegame_string  - decode a savegame string generated by build_savegame_string

Defined Exceptions:
    CanNotSaveGameException   -  raised when an object could not be encoded
    InvalidSaveGameException  -  raised when a savegame string is not trusted or could not be decoded

Example usage:
    # saving the game
    import savegame_codec  # load package
    savegame_string = savegame_codec.build_savegame_string()

    # loading the game
    import savegame_codec
    aistate = savegame_codec.load_savegame_string(savegame_string)
"""

from ._decoder import decode, load_savegame_string
from ._definitions import CanNotSaveGameException, InvalidSaveGameException
from ._encoder import build_savegame_string, encode



from inspect import getdoc


class FunctionInfo:
    def __init__(self, name: str, doc: str):
        self.name = name
        self.doc = doc


def inspect_function(name, value) -> FunctionInfo:
    return FunctionInfo(name, getdoc(value))



from focs._effects import EffectsGroup, IsBuilding, NamedReal, Planet, SetStealth, Value
from focs._tech import *

Tech(
    name="SPY_PLANET_STEALTH_MOD",
    description="SPY_PLANET_STEALTH_MOD_DESC",
    short_description="GAME_MAKE_WORK_SHORT_DESC",
    category="SPY_CATEGORY",
    researchcost=1,
    researchturns=1,
    researchable=False,
    tags=["PEDIA_SPY_CATEGORY"],
    effectsgroups=[
        EffectsGroup(
            scope=IsBuilding() | Planet(),
            stackinggroup="PLANET_STEALTH_MOD_STACK",
            accountinglabel="BASIC_PLANET_STEALTH",
            effects=SetStealth(value=Value + NamedReal(name="PLANET_BASIC_STEALTH", value=5.0)),
        )
    ],
    graphic="",
)



import freeOrionAIInterface as fo
from collections.abc import Mapping
from typing import Callable

import AIDependencies
from common.fo_typing import PlanetId, SpeciesName, SystemId
from freeorion_tools import ReadOnlyDict
from freeorion_tools.caching import cache_for_current_turn


class PlanetInfo:
    def __init__(self, pid: PlanetId):
        universe = fo.getUniverse()
        planet = universe.getPlanet(pid)
        self.pid: PlanetId = planet.id
        self.species_name = planet.speciesName
        self.owner = planet.owner
        self.system_id: SystemId = planet.systemID

    def is_colonized(self) -> bool:
        return self.owner == fo.empireID() and self.species_name

    def is_owned_by_empire(self) -> bool:
        return self.owner == fo.empireID()

    def has_outpost(self) -> bool:
        return self.owner == fo.empireID() and not self.species_name

    def is_not_owned(self) -> bool:
        return self.owner == AIDependencies.INVALID_ID and not self.species_name


@cache_for_current_turn
def _get_planets_info() -> Mapping[PlanetId, PlanetInfo]:
    universe = fo.getUniverse()
    planets = (universe.getPlanet(pid) for pid in universe.planetIDs)
    return {planet.id: PlanetInfo(planet.id) for planet in planets}


def _get_system_planets_map(planet_filter: Callable[[PlanetInfo], bool]) -> Mapping[SystemId, tuple[PlanetId]]:
    result = {}
    for planet_info in (planet_info for planet_info in _get_planets_info().values() if planet_filter(planet_info)):
        result.setdefault(planet_info.system_id, []).append(planet_info.pid)
    return ReadOnlyDict({k: tuple(v) for k, v in result.items()})


@cache_for_current_turn
def get_owned_planets() -> Mapping[SystemId, tuple[PlanetId]]:
    """
    Return map from system id to list of planet ids with colony or outpost.
    """
    return _get_system_planets_map(PlanetInfo.is_owned_by_empire)


@cache_for_current_turn
def get_colonized_planets() -> Mapping[SystemId, tuple[PlanetId]]:
    """
    Return map from system id to list of planet ids with colony only.
    """
    return _get_system_planets_map(PlanetInfo.is_colonized)


def get_owned_planets_in_system(sys_id: SystemId) -> tuple[PlanetId]:
    """
    Return list of planet ids with colony or outpost in the system.
    """
    return get_owned_planets().get(sys_id, ())


def get_colonized_planets_in_system(sys_id: SystemId) -> tuple[PlanetId]:
    """
    Return list of planets with colony in the system.
    """
    return get_colonized_planets().get(sys_id, ())


@cache_for_current_turn
def get_inhabited_planets() -> frozenset[PlanetId]:
    """
    Return frozenset of empire planet ids with species.
    """
    return frozenset(planet_info.pid for planet_info in _get_planets_info().values() if planet_info.is_colonized())


@cache_for_current_turn
def get_empire_outposts() -> tuple[PlanetId]:
    all_planets = _get_planets_info().values()
    return tuple(planet_info.pid for planet_info in all_planets if planet_info.has_outpost())


@cache_for_current_turn
def get_all_empire_planets() -> tuple[PlanetId]:
    all_planets = _get_planets_info().values()
    return tuple(planet_info.pid for planet_info in all_planets if planet_info.is_owned_by_empire())


@cache_for_current_turn
def get_empire_planets_by_species() -> Mapping[SpeciesName, list[PlanetId]]:
    """
    Return dict for empire from species to list of planet ids.
    """
    result = {}
    colonized = (planet_info for planet_info in _get_planets_info().values() if planet_info.is_colonized())
    for planet in colonized:
        result.setdefault(planet.species_name, []).append(planet.pid)
    return result


def get_unowned_empty_planets() -> set[PlanetId]:
    """
    Return the set of planets that are not owned by any player and have no natives.
    """
    return {planet_info.pid for planet_info in _get_planets_info().values() if planet_info.is_not_owned()}


def get_number_of_colonies() -> int:
    return len(get_inhabited_planets())


def get_empire_planets_with_species(species_name: str) -> tuple[PlanetId]:
    """
    Return empire planet ids with species.
    """
    if not species_name:
        return tuple()
    return tuple(get_empire_planets_by_species().get(species_name, []))



from collections import Counter, OrderedDict, defaultdict

import streamlit as st

from tabs.widgets import DIFF_ADD, DIFF_REMOVE, colored_span, span_with_hint, to_hex_color


def gather_policies_data(ais_data):
    data = OrderedDict()
    existing_empires = {}
    summary_stats = defaultdict(Counter)

    for empire in ais_data:
        empire_id = empire["empire_id"]
        existing_empires[empire_id] = empire["color"]
        for turn in empire["turns"]:
            all_added = turn.get("PolicyAdoption", [])
            summary_stats[empire_id].update(Counter(all_added))
            all_removed = turn.get("PolicyDeAdoption", [])

            summary_stats[empire_id].update({k: -v for k, v in Counter(all_removed).items()})

            # policies may be added or removed more than once per turn
            added_or_removed = {}
            for added in all_added:
                added_or_removed[added] = added_or_removed.get(added, 0) + 1
            for removed in all_removed:
                added_or_removed[removed] = added_or_removed.get(removed, 0) - 1
            added = sorted([x for x, value in added_or_removed.items() if value > 0])
            removed = sorted([x for x, value in added_or_removed.items() if value < 0])

            turn = turn["turn"]

            data.setdefault(turn, {})[empire_id] = (added, removed)

    return data, existing_empires, summary_stats


def norm_policy_name(name):
    name = name.split("PLC_", 1)[-1]
    return name.lower().capitalize()


def to_policy_span(text, color):
    return colored_span(span_with_hint(norm_policy_name(text), text), color)


def show_summary(existing_empires, summary_stats):
    st.header("Policies adopted")
    summary = []
    for empire_id, counts in summary_stats.items():
        counts = sorted([name for name, added in counts.items() if added])
        counts = ", ".join(to_policy_span(x, "") for x in counts)
        empire_id = colored_span(empire_id, to_hex_color(existing_empires[empire_id]))
        summary.append(f"- {empire_id}: {counts}")

    st.markdown("\n".join(summary), unsafe_allow_html=True)


def show_policies_table(data, existing_empires):
    header = " | ".join(colored_span(eid, to_hex_color(color)) for eid, color in sorted(existing_empires.items()))
    segments = " | ".join("---" for x in range(len(existing_empires) + 1))
    separator = f"| {segments} |"

    table = [
        f"| turn | {header} |",
        separator,
    ]

    is_gap = False

    for key, empires in data.items():
        has_changed = [bool(k) or bool(v) for k, v in empires.values()]
        if any(has_changed):
            is_gap = False

        else:
            if is_gap:
                continue
            else:
                is_gap = True

        def to_diff(val):
            added, removed = val

            return ", ".join(
                [*[to_policy_span(x, DIFF_ADD) for x in added], *[to_policy_span(x, DIFF_REMOVE) for x in removed]]
            )

        if is_gap:
            inner_row = " | ".join(["...", *[to_diff(v) for k, v in sorted(empires.items())]])
        else:
            inner_row = " | ".join([str(key), *[to_diff(v) for k, v in sorted(empires.items())]])

        table.append(f"| {inner_row} |")
    st.markdown("\n".join(table), unsafe_allow_html=True)


def plot_policy_adoptions(ais_data):
    data, existing_empires, summary_stats = gather_policies_data(ais_data)
    if not data:
        return

    show_policies_table(data, existing_empires)
    show_summary(existing_empires, summary_stats)



# This module should be imported with *, that means we should not import other things in it, otherwise things that should be imported from other modules will be imported from here.

from focs._types import _Condition, _EffectGroup, _PlanetEnvironment, _PlanetType

class _FocusType: ...

def FocusType(
    *,
    name: str,
    description: str,
    location: _Condition,
    graphic: str,
) -> _FocusType: ...
def Species(
    *,
    name: str,
    description: str,
    gameplay_description: str,
    tags: list[str],
    foci: list[_FocusType],
    defaultfocus: str,
    effectsgroups: list[_EffectGroup],
    graphic: str,
    environments: dict[_PlanetType, _PlanetEnvironment],
    likes: list[str] = [],
    dislikes: list[str] = [],
    can_produce_ships: bool = False,
    playable: bool = False,
    can_colonize: bool = False,
    native: bool = False,
    annexation_condition=None,
    spawnrate: int | None = None,
    spawnlimit: int | None = None,
    annexationcondition: _Condition | None = None,
    annexationcost: float = 0.0,
): ...
def SpeciesCensusOrdering(order: list[str]): ...



from operator import attrgetter

from stub_generator.interface_inspector import FunctionInfo
from stub_generator.parse_docs import Docs
from stub_generator.stub_generator.rtype import update_function_rtype


def _handle_function(fun: FunctionInfo):
    function = Docs(fun.doc, 1)
    return_annotation = " -> %s" % update_function_rtype(fun.name, function.rtype)
    docstring = function.get_doc_string()
    if docstring:
        docstring = f"\n{docstring}\n"
        end = ""
    else:
        end = " ..."
    arg_strings = list(function.get_argument_strings())
    if len(arg_strings) == 1:
        yield f"def {fun.name}({arg_strings[0]}){return_annotation}:{docstring}{end}"
    else:
        for arg_string in arg_strings:
            yield f"@overload\ndef {fun.name}({arg_string}){return_annotation}: ...\n"

        yield f"def {fun.name}(*args){return_annotation}:{docstring}{end}"


def generate_functions(functions: list[FunctionInfo]):
    for function in sorted(functions, key=attrgetter("name")):
        yield "".join(_handle_function(function))



# The SpeciesCensusOrdering.focs.txt file is used in
# default/scripting/species.  The format is a single line,

# `SpeciesCensusOrdering`

# followed by any number of lines like,

# `tag = "LITHIC"\n`

# The list of species tags provides the order that species are displayed
# and ordered in the empire Census pop-up.
from focs._species import *

SpeciesCensusOrdering(["LITHIC", "ORGANIC", "PHOTOTROPHIC", "ROBOTIC", "SELF_SUSTAINING", "TELEPATHIC", "GASEOUS"])



"""The FreeOrionAI module contains the methods which can be made by the C game client;
these methods in turn activate other portions of the python AI code."""
from functools import wraps
from logging import debug, error, fatal, info

from common.configure_logging import redirect_logging_to_freeorion_logger

# Logging is redirected before other imports so that import errors appear in log files.
redirect_logging_to_freeorion_logger()

import freeOrionAIInterface as fo
import random
import sys

from common.option_tools import parse_config

parse_config(fo.getOptionsDBOptionStr("ai-config"), fo.getUserConfigDir())

from freeorion_tools.fo_chat_handler import configure_debug_chat, process_chat_message
from freeorion_tools.patch_interface import patch_interface

patch_interface()

import ColonisationAI
import DiplomaticCorp
import ExplorationAI
import FleetUtilsAI
import InvasionAI
import MilitaryAI
import PlanetUtilsAI
import PolicyAI
import PriorityAI
import ProductionAI
import ResearchAI
import ResourcesAI
import ShipDesignAI
import TechsListsAI
from AIDependencies import INVALID_ID
from aistate_interface import create_new_aistate, get_aistate, load_aistate
from character.character_module import Aggression
from character.character_strings_module import (
    get_trait_name_aggression,
    possible_capitals,
)
from common.handlers import init_handlers
from common.listeners import listener
from empire.survey_universe import survey_universe
from expansion_plans.expansion_plans_implementation import initialise_expansion_plans
from freeorion_tools import chat_human
from freeorion_tools.timers import AITimer
from generate_orders import (
    empire_is_ok,
    greet_on_first_turn,
    print_existing_rules,
    print_starting_intro,
    replay_turn_after_load,
    set_game_turn_seed,
    update_resource_pool,
)

initialise_expansion_plans()

turn_timer = AITimer("full turn")

user_dir = fo.getUserDataDir()
debug("Path to folder for user specific data: %s" % user_dir)
debug("Python paths %s" % sys.path)


diplomatic_corp = None


def error_handler(func):
    """Decorator that logs any exception in decorated function, then re-raises"""

    @wraps(func)
    def _error_handler(*args, **kwargs):
        try:
            res = func(*args, **kwargs)
            return res
        except Exception as e:
            error("Exception %s occurred during %s", e, func.__name__, exc_info=True)
            raise

    return _error_handler


def _pre_game_start(empire_id):
    """
    Configuration that should be done before AI start operating.
    """
    aistate = get_aistate()
    aggression_trait = aistate.character.get_trait(Aggression)
    diplomatic_corp_configs = {
        fo.aggression.beginner: DiplomaticCorp.BeginnerDiplomaticCorp,
        fo.aggression.maniacal: DiplomaticCorp.ManiacalDiplomaticCorp,
    }
    global diplomatic_corp
    diplomatic_corp = diplomatic_corp_configs.get(aggression_trait.key, DiplomaticCorp.DiplomaticCorp)()
    TechsListsAI.test_tech_integrity()
    configure_debug_chat(empire_id)


def _choose_aggression():
    galaxy_setup_data = fo.getGalaxySetupData()
    aggression = int(galaxy_setup_data.maxAIAggression)

    if aggression > 0:
        rng = random.Random()
        galaxy_seed = hash(galaxy_setup_data.seed)
        empire_seed = 3 * hash(fo.getEmpire().name)
        rng.seed(galaxy_seed * empire_seed)

        random_number = rng.randint(0, 99)
        if random_number > 74:
            aggression -= 1

    return fo.aggression(aggression)


@error_handler
def startNewGame():  # pylint: disable=invalid-name
    """Called by client when a new game is started (but not when a game is loaded).
    Should clear any pre-existing state and set up whatever is needed for AI to generate orders."""
    empire = fo.getEmpire()
    if empire is None:
        fatal("This client has no empire. Ignoring new game start message.")
        return

    if empire.eliminated:
        info("This empire has been eliminated. Ignoring new game start message.")
        return
    # initialize AIstate
    debug("Initializing AI state...")
    aistate = create_new_aistate(_choose_aggression())
    aggression_trait = aistate.character.get_trait(Aggression)
    debug(
        "New game started, AI Aggression level %d (%s)"
        % (aggression_trait.key, get_trait_name_aggression(aistate.character))
    )
    aistate.session_start_cleanup()
    debug("Initialization of AI state complete!")
    debug("Trying to rename our homeworld...")
    planet_id = PlanetUtilsAI.get_capital()
    universe = fo.getUniverse()
    if planet_id != INVALID_ID:
        planet = universe.getPlanet(planet_id)
        new_name = " ".join([random.choice(possible_capitals(aistate.character)).strip(), planet.name])
        debug("    Renaming to %s..." % new_name)
        res = fo.issueRenameOrder(planet_id, new_name)
        debug("    Result: %d; Planet is now named %s" % (res, planet.name))
    _pre_game_start(empire.empireID)


@error_handler
def resumeLoadedGame(saved_state_string):  # pylint: disable=invalid-name
    """Called by client to when resume a loaded game."""
    debug("Resuming loaded game")

    if saved_state_string == "NO_STATE_YET" and fo.currentTurn() == 1:
        info("AI given uninitialized state-string to resume from on turn 1.")
        info(
            "Assuming post-universe-generation autosave before any orders were sent "
            "and behaving as if a new game was started."
        )
        return startNewGame()

    if fo.getEmpire() is None:
        fatal("This client has no empire. Doing nothing to resume loaded game.")
        return

    if fo.getEmpire().eliminated:
        info("This empire has been eliminated. Ignoring resume loaded game.")
        return

    aistate = None
    if saved_state_string == "NOT_SET_BY_CLIENT_TYPE":
        info("AI assigned to empire previously run by human.")
        chat_human("We have been assigned an empire previously run by a human player. We can manage this.")
    elif saved_state_string == "":
        error(
            "AI given empty state-string to resume from. "
            "AI can continue but behaviour may be different from the previous session."
        )
    else:
        try:
            # loading saved state
            aistate = load_aistate(saved_state_string)
        except Exception as e:
            error(
                "Failed to load the AIstate from the savegame: %s"
                " AI can continue but behaviour may be different from the previous session.",
                e,
                exc_info=True,
            )
    if aistate is None:
        info("Creating new ai state due to failed load.")
        aistate = create_new_aistate(_choose_aggression())
    aistate.session_start_cleanup()
    _pre_game_start(fo.getEmpire().empireID)


@error_handler
def prepareForSave():  # pylint: disable=invalid-name
    """Called by client when the game is about to be saved, to let the Python AI know it should save any AI state
    information, such as plans or knowledge about the game from previous turns,
    in the state string so that they can be restored if the game is loaded."""
    empire = fo.getEmpire()
    if empire is None:
        fatal("This client has no empire. Doing nothing to prepare for save.")
        return

    if empire.eliminated:
        info("This empire has been eliminated. Save info request")
        return

    info("Preparing for game save by serializing state")

    # serialize (convert to string) global state dictionary and send to AI client to be stored in save file
    import savegame_codec

    try:
        dump_string = savegame_codec.build_savegame_string()
        fo.setSaveStateString(dump_string)
    except Exception as e:
        error(
            "Failed to encode the AIstate as save-state string. "
            "The resulting save file should be playable but the AI "
            "may have a different aggression. The error raised was: %s" % e,
            exc_info=True,
        )


@error_handler
def handleChatMessage(sender_id, message_text):  # pylint: disable=invalid-name
    """Called when this player receives a chat message. sender_id is the player who sent the message, and
    message_text is the text of the sent message."""
    empire = fo.getEmpire()
    if empire is None:
        fatal("This client has no empire. Doing nothing to handle chat message.")
        return

    if empire.eliminated:
        debug("This empire has been eliminated. Ignoring chat message")
        return
    process_chat_message(sender_id, message_text, diplomatic_corp)


@error_handler
def handleDiplomaticMessage(message):  # pylint: disable=invalid-name
    """Called when this player receives a diplomatic message update from the server,
    such as if another player declares war, accepts peace, or cancels a proposed peace treaty."""
    empire = fo.getEmpire()
    if empire is None:
        fatal("This client has no empire. Doing nothing to handle diplomatic message.")
        return

    if empire.eliminated:
        debug("This empire has been eliminated. Ignoring diplomatic message")
        return

    diplomatic_corp.handle_diplomatic_message(message)


@error_handler
def handleDiplomaticStatusUpdate(status_update: fo.diplomaticStatusUpdate):  # pylint: disable=invalid-name
    """Called when this player receives an update about the diplomatic status between players, which may
    or may not include this player."""
    empire = fo.getEmpire()
    if empire is None:
        fatal("This client has no empire. Doing nothing to handle diplomatic status message.")
        return

    if empire.eliminated:
        debug("This empire has been eliminated. Ignoring diplomatic status update")
        return

    diplomatic_corp.handle_diplomatic_status_update(status_update)


generate_order_timer = AITimer("generate orders")


@error_handler
@listener
def generateOrders():  # pylint: disable=invalid-name
    """
    Called once per turn to tell the Python AI to generate
    and issue orders, i.e. to control its empire.

    After leaving this function, the AI's turn will be finished
    and its orders will be sent to the server.
    """
    turn_timer.start("AI planning")
    print_existing_rules()

    if not empire_is_ok():
        return

    DiplomaticCorp.check_gang_up()

    set_game_turn_seed()

    generate_order_timer.start("Update states on server")
    # This code block is required for correct AI work.
    update_resource_pool()

    generate_order_timer.start("Prepare each turn data")
    # results of this function are needed in many places...

    greet_on_first_turn(diplomatic_corp)

    if replay_turn_after_load():
        return

    survey_universe()

    print_starting_intro()

    aistate = get_aistate()
    aistate.prepare_for_new_turn()
    debug("Calling AI Modules")
    # call AI modules
    action_list = [
        ShipDesignAI.Cache.update_for_new_turn,
        PriorityAI.calculate_priorities,
        ExplorationAI.assign_scouts_to_explore_systems,
        ColonisationAI.assign_colony_fleets_to_colonise,
        InvasionAI.assign_invasion_fleets_to_invade,
        MilitaryAI.assign_military_fleets_to_systems,
        FleetUtilsAI.generate_fleet_orders_for_fleet_missions,
        FleetUtilsAI.issue_fleet_orders_for_fleet_missions,
        ResearchAI.generate_research_orders,
        ProductionAI.generate_production_orders,
        ResourcesAI.generate_resources_orders,
        PolicyAI.generate_policy_orders,
    ]

    for action in action_list:
        try:
            generate_order_timer.start(action.__name__)
            action()
            generate_order_timer.stop()
        except Exception as e:
            error(f"Exception {e} while trying to {action.__name__}", exc_info=True)

    aistate.last_turn_played = fo.currentTurn()
    generate_order_timer.stop_print_and_clear()
    turn_timer.stop_print_and_clear()
    turn_timer.start("Time between AI turn")


init_handlers(fo.getOptionsDBOptionStr("ai-config"), fo.getAIDir())



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="CON_CONC_CAMP",
    description="CON_CONC_CAMP_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="CONSTRUCTION_CATEGORY",
    researchcost=80 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_CONSTRUCTION_CATEGORY"],
    prerequisites=["CON_ARCH_PSYCH"],
    unlock=Item(type=UnlockBuilding, name="BLD_CONC_CAMP"),
    graphic="icons/building/concentration-camp.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="GRO_GENETIC_ENG",
    description="GRO_GENETIC_ENG_DESC",
    short_description="THEORY_SHORT_DESC",
    category="GROWTH_CATEGORY",
    researchcost=9 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_GROWTH_CATEGORY", "THEORY"],
    graphic="icons/tech/genetic_engineering.png",
)



import freeOrionAIInterface as fo
from typing import Any

from freeorion_tools import chat_human
from freeorion_tools.chat_handler.base_chat_handler import ChatHandlerBase
from freeorion_tools.chat_handler.chat_formatter import ChatFormatter
from freeorion_tools.chat_handler.interpreter import DebugInterpreter
from freeorion_tools.chat_handler.shell_variable import ShellVariable


def _chat_printer(message: Any):
    return chat_human(message, send_to_logs=False)


debug_chat_handler_local_vars = _shell_locals = [
    ShellVariable(
        variable="fo",
        expression="freeOrionAIInterface",
        description="fo",
        imports=("import freeOrionAIInterface",),
    ),
    ShellVariable(
        variable="ai",
        expression="get_aistate()",
        description="aistate",
        imports=("from aistate_interface import get_aistate",),
    ),
    ShellVariable(
        variable="u",
        expression="fo.getUniverse()",
        description="universe",
    ),
    ShellVariable(
        variable="e",
        expression="fo.getEmpire()",
        description="empire",
    ),
]


class Provider:
    def get_player_ids(self):
        return fo.allPlayerIDs()

    def is_host(self, player_id):
        return fo.playerIsHost(player_id)

    def player_id(self):
        return fo.playerID()

    def player_name(self, player_id):
        return fo.playerName(player_id)

    def get_empire_id(self, player_id):
        empire = fo.getEmpire(fo.playerEmpireID(player_id))
        return empire.empireID

    def get_empire_name(self, player_id):
        empire = fo.getEmpire(fo.playerEmpireID(player_id))
        return empire.name

    def get_empire_color(self, player_id):
        empire = fo.getEmpire(fo.playerEmpireID(player_id))
        return empire.colour


class DebugChatHandler(ChatHandlerBase):
    """
    Handle console messages to provide debug chat.

    This handler allows executing arbitrary python code for the AIs players.
    Mainly it is supposed to be used for debugging purposed, but can be used for cheating too.
    """

    _entering_debug_message = "Entering debug mode"
    _exit_debug_message = "Exiting debug mode"

    def __init__(
        self,
        empire_id: int,
        chat_printer=_chat_printer,
        log_printer=print,
        formatter=ChatFormatter(),
        shell_locals=None,
        provider=None,
    ):
        super().__init__()
        if not shell_locals:
            shell_locals = debug_chat_handler_local_vars

        if not provider:
            provider = Provider()
        self._provider = provider
        self._shell_locals = shell_locals
        self._formatter = formatter
        self._empire_id = empire_id
        self._debug_mode = False
        self._interpreter = DebugInterpreter(self._shell_locals)
        self._log_printer = log_printer
        self._chat_printer = chat_printer

    def send_notification(
        self,
        chat_message: Any = None,
        log_message: Any = None,
    ):
        """Handle user notifications wia logs and chat."""
        if chat_message is not None:
            self._chat_printer(chat_message)
        if log_message is not None:
            self._log_printer(log_message)

    def process_message(self, sender_id: int, message: str) -> bool:
        host_id = [x for x in self._provider.get_player_ids() if self._provider.is_host(x)][0]

        if sender_id != host_id:
            return False  # only user can send debug message

        if self._debug_mode:
            return self._handle_debug_commands(message)
        else:
            return self._handle_chat_commands(message)

    def _handle_chat_commands(self, message):
        if message.startswith("start"):
            self._handle_start(message)
        elif message == "help" and not self._debug_mode:
            self._handle_help()
        elif message == "stop":
            self.debug_active = False
        else:
            return False
        return True

    def _handle_debug_commands(self, message):
        if message == "stop":
            self._handle_stop()
        else:
            self._handle_shell_input(message)
        return True

    # message handlers
    def _handle_stop(self):
        self.send_notification(
            chat_message=self._exit_debug_message,
            log_message=self._exit_debug_message,
        )
        self.debug_active = False
        self._debug_mode = False

    def _handle_shell_input(self, message):
        """Handle message and prints it to chat and logs."""
        self.send_notification(log_message=f">>> {message}")
        out, err = self._interpreter.eval(message)
        if out:
            self.send_notification(
                chat_message=self._formatter.white(out),
                log_message=out,
            )
        if err:
            self.send_notification(
                chat_message=self._formatter.red(err),
                log_message=err,
            )

    def _handle_start(self, message):
        try:
            player_id = int(message[5:].strip())
        except ValueError:
            if self._ai_should_respond():
                text = "Invalid empire id, please input valid number"
                self.send_notification(
                    chat_message=self._formatter.red(text),
                )
                self._handle_help()
            return
        self.debug_active = True
        # This message is not for that AI
        if player_id != self._provider.player_id():
            return
        self._start_debug_mode()

    def _start_debug_mode(self):
        self._debug_mode = True

        start_message = [
            self._entering_debug_message,
            "Print %s to exit." % self._formatter.blue("stop"),
            "Local variables:",
        ]

        variables_description = self._interpreter.set_locals()
        for var, name in variables_description:
            start_message.append(
                "  {}{}{}".format(
                    self._formatter.underline(self._formatter.yellow(var)),
                    " " * (3 - len(var)),  # cannot use normal formatting because of tags
                    name,
                )
            )
        self.send_notification(chat_message="\n".join(self._formatter.white(x) for x in start_message))

    def _ai_should_respond(self):
        """
        Return true if AI should respond to message.

        To avoid chat pollution only one AI should respond to the message.
        AI player with the smallest id is selected.
        """
        ais = [x for x in self._provider.get_player_ids() if not self._provider.is_host(x)]
        return ais[0] == self._provider.player_id()

    def _handle_help(self):
        if not self._ai_should_respond():
            return

        help_message = [
            self._formatter.white("Chat commands:"),
            self._formatter.white(
                "  "
                + self._formatter.underline(self._formatter.blue("start <id>"))
                + ": start debug for selected empire"
            ),
        ]
        for player_id in self._provider.get_player_ids():
            if not self._provider.is_host(player_id):
                help_message.append(self._get_empire_string(player_id))
        self._formatter.white("  " + self._formatter.underline(self._formatter.blue("stop")) + ": stop debug"),
        self.send_notification(chat_message="\n".join(help_message))

    def _get_empire_string(self, player_id):
        name = self._provider.get_empire_name(player_id)
        color = self._provider.get_empire_color(player_id)

        return (
            "    "
            + self._formatter.underline(self._formatter.blue(player_id))
            + self._formatter.colored(color, " %s" % name)
            + self._formatter.white(" by %s" % self._provider.player_name(player_id))
        )



import freeOrionAIInterface as fo

from common.option_tools import get_option_dict
from DiplomaticCorp import DiplomaticCorp, handle_pregame_chat
from freeorion_tools.chat_handler import DebugChatHandler, NormalChatHandler

_chat_handler = None  # Optional[_ChatHandler]


def configure_debug_chat(empire_id: int) -> None:
    global _chat_handler
    if get_option_dict().get("allow_debug_chat", False):
        _chat_handler = DebugChatHandler(empire_id)
    else:
        _chat_handler = NormalChatHandler()


def process_chat_message(sender_id: None, message_text: str, diplomatic_corp: DiplomaticCorp):
    # We got our first massages before we initialize AI
    if _chat_handler is None:
        handle_pregame_chat(sender_id, message_text)
        return
    handled = _chat_handler.process_message(sender_id, message_text)
    if not handled and not (fo.playerIsHost(sender_id) and _chat_handler.debug_active):
        diplomatic_corp.handle_midgame_chat(sender_id, message_text)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_DARK_RAY",
    description="SHP_DARK_RAY_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=120 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_BOMBARD_WEAPON_TECHS"],
    prerequisites=["SHP_BOMBARD", "SPY_STEALTH_2"],
    unlock=Item(type=UnlockShipPart, name="SP_DARK_RAY"),
    graphic="icons/ship_parts/dark-ray.png",
)



"""col_bld_gen.py

Utility script to generate consistent colony building definitions for each species.
This script is not utilized by the game.

When executed, definition files will be generated in the current working directory.
"""

import os
import os.path
import string
from itertools import chain

# List of all colonizable species in game: definition key, graphic file(relative to default/data/art)
from typing import Any

species_list = [
    ("SP_SUPER_TEST", "icons/species/other-04.png"),
    ("SP_ABADDONI", "icons/species/abaddonnian.png"),
    ("SP_BANFORO", "icons/species/banforo.png"),
    ("SP_CHATO", "icons/species/chato-matou-gormoshk.png"),
    ("SP_CRAY", "icons/species/cray.png"),
    ("SP_DERTHREAN", "icons/species/derthrean.png"),
    ("SP_EAXAW", "icons/species/eaxaw.png"),
    ("SP_EGASSEM", "icons/species/egassem.png"),
    ("SP_ETTY", "icons/species/etty.png"),
    ("SP_FULVER", "icons/species/insectoid-01.png"),
    ("SP_FURTHEST", "icons/species/furthest.png"),
    ("SP_GEORGE", "icons/species/george.png"),
    ("SP_GYSACHE", "icons/species/gysache.png"),
    ("SP_HAPPY", "icons/species/ichthyoid-06.png"),
    ("SP_HHHOH", "icons/species/hhhoh.png"),
    ("SP_HUMAN", "icons/species/human.png"),
    ("SP_KILANDOW", "icons/species/kilandow.png"),
    ("SP_KOBUNTURA", "icons/species/intangible-04.png"),
    ("SP_LAENFA", "icons/species/laenfa.png"),
    ("SP_MISIORLA", "icons/species/misiorla.png"),
    ("SP_MUURSH", "icons/species/muursh.png"),
    ("SP_PHINNERT", "icons/species/phinnert.png"),
    ("SP_SCYLIOR", "icons/species/scylior.png"),
    ("SP_SETINON", "icons/species/amorphous-02.png"),
    ("SP_SILEXIAN", "icons/species/robotic-06.png"),
    ("SP_SLY", "icons/species/amorphous-05.png"),
    ("SP_SSLITH", "icons/species/sslith.png"),
    ("SP_TAEGHIRUS", "icons/species/t-aeghirus.png"),
    ("SP_TRITH", "icons/species/trith.png"),
    ("SP_REPLICON", "icons/species/replicon.png"),
    ("SP_UGMORS", "icons/species/amorphous-06.png"),
    ("SP_EXOBOT", "icons/species/robotic-01.png"),
]

# Species which start as extinct and tech enabling colonization without a suitable colony in supply range
species_extinct_techs = {
    "SP_BANFORO": "TECH_COL_BANFORO",
    "SP_KILANDOW": "TECH_COL_KILANDOW",
    "SP_MISIORLA": "TECH_COL_MISIORLA",
}

# Default gamerule to toggle availablity of colony buildings for a species (no rule)
colony_gamerule_default = ""

# Species specific gamerule enabling colony building
species_colony_gamerules = {"SP_SUPER_TEST": "RULE_ENABLE_SUPER_TESTER"}

# default base buildcost
buildcost_default = 50

# Species specific overrides to base buildcost
species_buildcost = {"SP_EXOBOT": 60}

# default buildtime factor
buildtime_factor_default = "1.0"

# Species specific overrides to default buildtime factor
species_time_factor = {"SP_HAPPY": "1.2", "SP_PHINNERT": "0.75"}


# Main template
t_main = string.Template(
    """// For long term changes - Do not modify this definition directly
//                     Instead modify and execute col_bld_gen.py and use the result.
BuildingType
    name = "BLD_COL_${name}"
    description = "BLD_COL_${name}_DESC"
    buildcost = ${cost} * [[COLONY_UPKEEP_MULTIPLICATOR]] * [[BUILDING_COST_MULTIPLIER]] * [[COLONIZATION_POLICY_MULTIPLIER]]
    buildtime = ${time}
    tags = [ ${tags} ]
    location = And [
        Planet
        OwnedBy empire = Source.Owner
        Population high = 0
        Not Planet environment = Uninhabitable species = "${id}"
        ${species_condition}
    ]
    enqueuelocation = And [
        Planet
        OwnedBy empire = Source.Owner
        Population high = 0
        Not Planet environment = Uninhabitable species = "${id}"
        ${exclude_parallel_colonies}
        ${species_condition}
    ]
    effectsgroups = [
        [[LIFECYCLE_MANIP_POPULATION_EFFECTS("${id}")]]

        EffectsGroup
            scope = And [
                Object id = Source.PlanetID
                Planet
            ]
            activation = Turn low = Source.CreationTurn + 1 high = Source.CreationTurn + 1
            effects = [
                GenerateSitRepMessage
                    message = "SITREP_NEW_COLONY_ESTABLISHED"
                    label = "SITREP_NEW_COLONY_ESTABLISHED_LABEL"
                    icon = "${graphic}"
                    parameters = [
                        tag = "species" data = "${id}"
                        tag = "planet" data = Target.ID
                    ]
                    empire = Source.Owner
            ]

        EffectsGroup
            scope = Source
            activation = Turn low = Source.CreationTurn + 2
            effects = Destroy
    ]
    icon = "${graphic}"

#include "/scripting/macros/misc.macros"
#include "/scripting/macros/upkeep.macros"
#include "/scripting/macros/priorities.macros"
#include "/scripting/macros/base_prod.macros"
#include "/scripting/macros/misc.macros"
"""
)

# Location and Enqueued condition template
t_species_cond = string.Template(
    """ResourceSupplyConnected empire = Source.Owner condition = And [
            Planet
            OwnedBy empire = Source.Owner
            Species name = "${id}"
            Population low = [[MIN_RECOLONIZING_SIZE]]
            Happiness low = 5
        ]"""
)

# Location and Enqueued condition template for extinct species
t_species_cond_extinct = string.Template(
    """ResourceSupplyConnected empire = Source.Owner condition = And [
            Planet
            OwnedBy empire = Source.Owner
            Or [
                And [
                    Species name = "${id}"
                    Population low = [[MIN_RECOLONIZING_SIZE]]
                    Happiness low = 5
                ]
                And [
                    OwnerHasTech name = "${tech_name}"
                    HasSpecial name = "EXTINCT_${name}_SPECIAL"
                    Contains Building name = "BLD_XENORESURRECTION_LAB"
                ]
            ]
        ]"""
)

# buildtime statistic condition template
t_buildtime_stat_cond = string.Template(
    """condition = And [
                Planet
                OwnedBy empire = Source.Owner
                Species name = "${id}"
                Population low = [[MIN_RECOLONIZING_SIZE]]
                Happiness low = 5
                ResourceSupplyConnected empire = Source.Owner condition = Target
            ]"""
)

# buildtime statistic condition template for extinct species
t_buildtime_stat_cond_extinct = string.Template(
    """condition = And [
                Planet
                OwnedBy empire = Source.Owner
                Or [
                   And [
                       Species name = "${id}"
                       Population low = [[MIN_RECOLONIZING_SIZE]]
                       Happiness low = 5
                    ]
                    And [
                        HasSpecial name = "EXTINCT_${name}_SPECIAL"
                        Contains Building name = "BLD_XENORESURRECTION_LAB"
                    ]
                ]
                ResourceSupplyConnected empire = Source.Owner condition = Target
            ]"""
)

# buildtime template
t_buildtime = string.Template(
    """${t_factor} * max(5.0, 1.0 +
        (Statistic Min value = ShortestPath object = Target.SystemID object = LocalCandidate.SystemID
            ${stat_condition}
        ) / (60
             + 20 * (Statistic If condition = Or [
                 And [ Source OwnerHasTech name = "SHP_MIL_ROBO_CONT" ]
                 And [ Source OwnerHasTech name = "SHP_SPACE_FLUX_BUBBLE" ]
                 And [ Source OwnerHasTech name = "SHP_ORG_HULL" ]
                 And [ Source OwnerHasTech name = "SHP_QUANT_ENRG_MAG" ]
             ])
             + 20 * (Statistic If condition = Or [
                 And [ Source OwnerHasTech name = "SHP_ORG_HULL" ]
                 And [ Source OwnerHasTech name = "SHP_QUANT_ENRG_MAG" ]
             ])
             + 20 * (Statistic If condition = And [ Source OwnerHasTech name = "SHP_QUANT_ENRG_MAG" ])
             + 10 * (Statistic If condition = And [ Source OwnerHasTech name = "SHP_IMPROVED_ENGINE_COUPLINGS" ])
             + 10 * (Statistic If condition = And [ Source OwnerHasTech name = "SHP_N_DIMENSIONAL_ENGINE_MATRIX" ])
             + 10 * (Statistic If condition = And [ Source OwnerHasTech name = "SHP_SINGULARITY_ENGINE_CORE" ])
             + 10 * (Statistic If condition = And [ Source OwnerHasTech name = "SHP_TRANSSPACE_DRIVE" ])
             + 10 * (Statistic If condition = And [ Source OwnerHasTech name = "SHP_INTSTEL_LOG" ])
        )
    )"""
)

exclude_parallel_colonies = ""
for species in chain((x[0] for x in species_list), species_extinct_techs.keys()):
    name = species[3:]
    exclude_parallel_colonies += f"""\
        Not Contains And [ Building name = "BLD_COL_{name}" OwnedBy empire = Source.Owner ]
        Not Enqueued type = Building name = "BLD_COL_{name}"
"""
# remove indent from first line and newline at the end to match format expected by the template
exclude_parallel_colonies = exclude_parallel_colonies[8:-1]

outpath = os.getcwd()
print("Output folder: %s" % outpath)

for sp_id, sp_graphic in species_list:
    sp_name = sp_id.split("_", 1)[1]
    sp_tags = '"' + sp_id + '"'
    sp_filename = sp_id + ".focs.txt"
    sp_gamerule = species_colony_gamerules.get(sp_id, colony_gamerule_default)
    extinct_tech = species_extinct_techs.get(sp_id, "")

    data: dict[str, Any] = {
        "id": sp_id,
        "name": sp_name,
        "tags": sp_tags,
        "graphic": sp_graphic,
        "cost": species_buildcost.get(sp_id, buildcost_default),
        "time": "",
        "exclude_parallel_colonies": exclude_parallel_colonies,
        "species_condition": "",
    }

    if sp_id == "SP_EXOBOT":
        data["tags"] += ' "CTRL_ALWAYS_REPORT"'
        data["time"] = 5
        data["species_condition"] = r"// no existing Exobot colony required!"
    else:
        if extinct_tech != "":
            data["tags"] += ' "CTRL_EXTINCT"'
            data["time"] = t_buildtime.substitute(
                t_factor=species_time_factor.get(sp_id, buildtime_factor_default),
                stat_condition=t_buildtime_stat_cond_extinct.substitute(id=sp_id, name=sp_name),
            )
            data["species_condition"] = t_species_cond_extinct.substitute(
                tech_name=extinct_tech, id=sp_id, name=sp_name
            )
        else:
            data["time"] = t_buildtime.substitute(
                t_factor=species_time_factor.get(sp_id, buildtime_factor_default),
                stat_condition=t_buildtime_stat_cond.substitute(id=sp_id),
            )
            data["species_condition"] = t_species_cond.substitute(id=sp_id)

    if sp_gamerule:
        data["species_condition"] += '\n        ((GameRule name = "%s") > 0)' % sp_gamerule

    with open(os.path.join(outpath, sp_filename), "w") as f:
        f.write(t_main.substitute(**data))



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import BROAD_EP, TOXIC_BROAD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import BAD_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.planet_size import LARGE_PLANET
from species.species_macros.population import BAD_POPULATION
from species.species_macros.research import GREAT_RESEARCH
from species.species_macros.shields import GREAT_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import BAD_DEFENSE_TROOPS, BAD_OFFENSE_TROOPS
from species.species_macros.weapons import BAD_WEAPONS

Species(
    name="SP_CHATO",
    description="SP_CHATO_DESC",
    gameplay_description="SP_CHATO_GAMEPLAY_DESC",
    playable=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "PHOTOTROPHIC",
        "GREAT_RESEARCH",
        "BAD_INFLUENCE",
        "BAD_WEAPONS",
        "GREAT_SHIP_SHIELDS",
        "BAD_POPULATION",
        "AVERAGE_SUPPLY",
        "BAD_OFFENSE_TROOPS",
        "PEDIA_PHOTOTROPHIC_SPECIES_CLASS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "FOCUS_RESEARCH",
        "FRACTAL_GEODES_SPECIAL",
        "SHIMMER_SILK_SPECIAL",
        "FRUIT_SPECIAL",
        "MONOPOLE_SPECIAL",
        "GAIA_SPECIAL",
        "ELERIUM_SPECIAL",
        "SPICE_SPECIAL",
        "PLC_CONFORMANCE",
        "PLC_DIVINE_AUTHORITY",
        "PLC_DREAM_RECURSION",
        "PLC_MODERATION",
    ],
    dislikes=[
        "BLD_PLANET_CLOAK",
        "BLD_SCANNING_FACILITY",
        "BLD_INTERSPECIES_ACADEMY",
        "BLD_SHIPYARD_CON_ADV_ENGINE",
        "BLD_NEUTRONIUM_FORGE",
        "BLD_STARGATE",
        "ECCENTRIC_ORBIT_SPECIAL",
        "KRAKEN_NEST_SPECIAL",
        "PANOPTICON_SPECIAL",
        "PLC_LIBERTY",
        "PLC_STOCKPILE_LIQUIDATION",
        "PLC_ALGORITHMIC_RESEARCH",
        "PLC_ARTISAN_WORKSHOPS",
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        *GREAT_RESEARCH,
        *BAD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *BAD_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_CHATO"),
        *AVERAGE_SUPPLY,
        *BAD_DEFENSE_TROOPS,
        *BAD_OFFENSE_TROOPS,
        *BAD_WEAPONS,
        # not for description
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *LARGE_PLANET,
        BROAD_EP,
        *GREAT_SHIP_SHIELDS,
    ],
    environments=TOXIC_BROAD_EP,
    graphic="icons/species/chato-matou-gormoshk.png",
)



from collections.abc import Collection
from typing import Any

from common.print_utils._base_field import Field


class Text(Field):
    def __init__(self, name, description="", align="<"):
        super().__init__(name, align=align, description=description)

    def convert_value_to_string(self, val: Any) -> str:
        return str(val)


class Number(Field):
    def __init__(self, name, precession=2, align=">", description="", placeholder="", total=False):
        super().__init__(
            name,
            align=align,
            description=description,
            placeholder=placeholder,
            total=total,
        )
        self.precession = precession

    def convert_value_to_string(self, val: Any) -> str:
        return "{: .{precession}f}".format(val, precession=self.precession)


class Bool(Field):
    def __init__(self, name, no_yes=("-", "+"), description=""):
        self.no_yes = no_yes
        assert len(no_yes) == 2
        super().__init__(name, description=description)

    def convert_value_to_string(self, val: Any) -> str:
        return self.no_yes[val]


class Sequence(Text):
    def convert_value_to_string(self, vals: Collection[Any]) -> str:
        return ", ".join(str(val) for val in vals)



from focs._effects import (
    EmpireHasAdoptedPolicy,
    IsSource,
    Source,
    StatisticIf,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="PRO_ORBITAL_GEN",
    description="PRO_ORBITAL_GEN_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=128
    * TECH_COST_MULTIPLIER
    * (
        1
        - 0.25
        * StatisticIf(float, condition=IsSource & EmpireHasAdoptedPolicy(empire=Source.Owner, name="PLC_INDUSTRIALISM"))
    ),
    researchturns=4,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    prerequisites=["PRO_FUSION_GEN", "CON_ORBITAL_CON"],
    unlock=Item(type=UnlockBuilding, name="BLD_GAS_GIANT_GEN"),
    graphic="icons/tech/orbital_generation.png",
)



"""
character_strings_module.py produces strings and lists of strings based on
Character for communicating with players and devs.

There are two types of public functions: functions that provide a readable name
of the level/type of a trait and functions that provide a list of strings to be
used as responses in the UI.

The functions of the type get_trait_name_X(character) return a user string
describing the nature of the trait X if the character has that trait or None if
the character does not have that trait.

For example:
get_trait_name_aggression(Character([Aggression(1), EmpireID(0,0)]])) returns "Turtle"
while
get_trait_name_aggression(Character([EmpireID(0,0)])) returns None

The functions of the type possible_X(character) return a list of strings based
on character suitable for the UI.

For example:
possible_capitals(Character([Aggression(0)])) returns ['Royal', 'Imperial'].
"""
import freeOrionAIInterface as fo
from logging import debug

import character as character_package


class _CharacterTableFunction:
    """A table indexed by a particular trait of a Character that is used like a
    function.

    When called this object will use the appropriate trait to lookup and
    possibly post process a line in the table.

    """

    def __init__(self, trait_class, table, post_process_func=None):
        """Store trait type, table and post processing function."""
        self.trait_class = trait_class
        self.table = table
        self.post_process = post_process_func

    def __repr__(self):
        return str(self.table)

    def __call__(self, character):
        """Return the indexed and post-processed string. Get the key from the correct trait in the character."""
        trait = character.get_trait(self.trait_class)
        if trait is None and None not in self.table:
            return None
        elem = self.table[trait.key] if trait is not None else self.table[None]
        if self.post_process:
            elem = self.post_process(elem)
        debug("CharacterTable returns %s", elem)
        return elem


_aggression_label_suffix = {
    fo.aggression.beginner: "_BEGINNER",
    fo.aggression.turtle: "_TURTLE",
    fo.aggression.cautious: "_CAUTIOUS",
    fo.aggression.typical: "_TYPICAL",
    fo.aggression.aggressive: "_AGGRESSIVE",
    fo.aggression.maniacal: "_MANIACAL",
}


def _make_aggression_based_function(prefix, post_process_func=None):
    """Make an aggression _CharacterTableFunction"""
    table = {key: f"{prefix}{suffix}" for (key, suffix) in _aggression_label_suffix.items()}
    table[None] = "UNKNOWN_VALUE_SYMBOL"
    return _CharacterTableFunction(character_package.character_module.Aggression, table, post_process_func)


# Human readable trait name functions for debugging
get_trait_name_aggression = _make_aggression_based_function("GSETUP", fo.userString)

# Lists of strings for the UI based on character
possible_capitals = _make_aggression_based_function("AI_CAPITOL_NAMES", fo.userStringList)
possible_greetings = _make_aggression_based_function("AI_FIRST_TURN_GREETING_LIST", fo.userStringList)



import freeOrionAIInterface as fo


# this name left with C naming style for compatibility with translation assistance procedures
def UserString(label, default=None):  # pylint: disable=invalid-name
    """
    A translation assistance tool is intended to search for this method to identify translatable strings.

    :param label: a UserString key
    :param default: a default value to return if there is a key error
    :return: a translated string for the label
    """

    table_string = fo.userString(label)

    if "ERROR: " + label in table_string:  # implement test for string lookup not found error
        return default or table_string
    else:
        return table_string


# this name left with C naming style for compatibility with translation assistance procedures
def UserStringList(label):  # pylint: disable=invalid-name
    """
    A translation assistance tool is intended to search for this method to identify translatable strings.

    :param label: a UserString key
    :return: a python list of translated strings from the UserString list identified by the label
    """

    return fo.userStringList(label)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="CON_PLANET_DRIVE",
    description="CON_PLANET_DRIVE_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="CONSTRUCTION_CATEGORY",
    researchcost=360 * TECH_COST_MULTIPLIER,
    researchturns=6,
    tags=["PEDIA_CONSTRUCTION_CATEGORY"],
    prerequisites=["LRN_SPATIAL_DISTORT_GEN"],
    unlock=[
        Item(type=UnlockBuilding, name="BLD_PLANET_DRIVE"),
        Item(type=UnlockBuilding, name="BLD_PLANET_BEACON"),
        Item(type=UnlockShipPart, name="SP_PLANET_BEACON"),
    ],
    graphic="icons/tech/planetary_stardrive.png",
)



import freeorion as fo
import random

import universe_statistics
import universe_tables
from galaxy import DisjointSets
from util import MapGenerationError, report_error


class StarlaneAlteringMonsters:
    def __init__(self, systems):
        self.systems = systems
        self.placed = set()
        self.removed_lanes = set()
        self.starlanes = [(s1, s2) for s1 in systems for s2 in fo.sys_get_starlanes(s1) if s1 < s2]

    def can_place_at(self, system, plan):
        """
        Check if the starlane altering monster fleet ''plan'' can be placed at ''system''
        without disjoining the galaxy map.
        Compute the disjoint set of the galaxy without the starlanes
        from the proposed system.  Return False if there will be more
        connected regions than the number of placed starlane altering
        monsters plus one, otherwise True.
        """
        local_lanes = {(min(system, s), max(system, s)) for s in fo.sys_get_starlanes(system)}

        dsets = DisjointSets()
        for s in self.systems:
            dsets.add(s)

        for lane in self.starlanes:
            if lane in self.removed_lanes or lane in local_lanes:
                continue
            dsets.link(lane[0], lane[1])

        num_contiguous_regions = len(dsets.complete_sets())
        expected_num_contiguous_regions = len(self.placed) + 2

        if num_contiguous_regions > expected_num_contiguous_regions:
            return False

        if num_contiguous_regions < expected_num_contiguous_regions:
            report_error(
                "Number of contiguous regions %d is below the expected number "
                "of contiguous regions %d when placing %d monster %s that can "
                "break starlanes."
                % (num_contiguous_regions, expected_num_contiguous_regions, len(self.placed) + 1, plan.name())
            )
            return False

        return True

    def place(self, system, plan):
        """
        Create map altering monster fleet ''plan'' at ''system''.
        """
        local_lanes = {(min(system, s), max(system, s)) for s in fo.sys_get_starlanes(system)}

        populate_monster_fleet(plan, system)
        if system not in self.placed:
            self.placed.add(system)
            self.removed_lanes |= local_lanes


def populate_monster_fleet(fleet_plan, system):
    """
    Create a monster fleet in ''system'' according to ''fleet_plan''
    """

    # create monster fleet
    monster_fleet = fo.create_monster_fleet(system)

    if monster_fleet == fo.invalid_object():
        raise MapGenerationError("Python generate_monsters: unable to create new monster fleet %s" % fleet_plan.name())

    # add monsters to fleet
    for design in fleet_plan.ship_designs():
        if fo.create_monster(design, monster_fleet) == fo.invalid_object():
            raise MapGenerationError("Python generate_monsters: unable to create monster %s" % design)

    print("Spawn", fleet_plan.name(), "at", fo.get_name(system))


def generate_monsters(monster_freq, systems):  # noqa: C901
    """
    Adds space monsters to systems.
    """
    # first, calculate the basic chance for monster generation in a system
    # based on the monster frequency that has been passed
    # get the corresponding value for the specified monster frequency from the universe tables
    basic_chance = universe_tables.MONSTER_FREQUENCY[monster_freq]
    # a value of 0 means no monsters, in this case return immediately
    if basic_chance <= 0:
        return
    print("Default monster spawn chance:", basic_chance)
    expectation_tally = 0.0
    actual_tally = 0

    # get all monster fleets that have a spawn rate and limit both > 0 and at least one monster ship design in it
    # (a monster fleet with no monsters in it is pointless) and store them in a list
    fleet_plans = fo.load_monster_fleet_plan_list()

    # create a map where we store a spawn counter for each monster fleet
    # this counter will be set to the spawn limit initially and decreased every time the monster fleet is spawned
    # this map (dict) needs to be separate from the list holding the fleet plans because the order in which items
    # are stored in a dict is undefined (can be different each time), which would result in different distribution
    # even when using the same seed for the RNG
    spawn_limits = {
        fp: fp.spawn_limit()
        for fp in fleet_plans
        if fp.spawn_rate() > 0.0 and fp.spawn_limit() > 0 and fp.ship_designs()
    }

    # map nests to monsters for ease of reporting
    nest_name_map = {
        "KRAKEN_NEST_SPECIAL": "SM_KRAKEN_1",
        "SNOWFLAKE_NEST_SPECIAL": "SM_SNOWFLAKE_1",
        "JUGGERNAUT_NEST_SPECIAL": "SM_JUGGERNAUT_1",
    }
    tracked_plan_tries = {name: 0 for name in nest_name_map.values()}
    tracked_plan_counts = {name: 0 for name in nest_name_map.values()}
    tracked_plan_valid_locations = {fp: 0 for fp in fleet_plans if fp.name() in tracked_plan_counts}

    if not fleet_plans:
        return

    universe = fo.get_universe()

    # Fleet plans that include ships capable of altering starlanes.
    # @content_tag{CAN_ALTER_STARLANES} universe_generator special handling
    # for fleets containing a hull design with this tag.
    fleet_can_alter_starlanes = {
        fp
        for fp in fleet_plans
        if any(
            [
                universe.getGenericShipDesign(design).ship_hull.hasTag("CAN_ALTER_STARLANES")
                for design in fp.ship_designs()
            ]
        )
    }

    # dump a list of all monster fleets meeting these conditions and their properties to the log
    print("Monster fleets available for generation at game start:")
    fp_location_cache = {}
    for fleet_plan in fleet_plans:
        print("...", fleet_plan.name(), ": spawn rate", fleet_plan.spawn_rate(), end=" ")
        print("/ spawn limit", fleet_plan.spawn_limit(), end=" ")
        print("/ effective chance", basic_chance * fleet_plan.spawn_rate(), end=" ")
        fp_location_cache[fleet_plan] = set(fleet_plan.locations(systems))
        print("/ can be spawned at", len(fp_location_cache[fleet_plan]), "of", len(systems), "systems")
        if fleet_plan.name() in nest_name_map.values():
            universe_statistics.tracked_monsters_chance[fleet_plan.name()] = basic_chance * fleet_plan.spawn_rate()

    # initialize a manager for monsters that can alter the map
    # required to prevent their placement from disjoining the map
    starlane_altering_monsters = StarlaneAlteringMonsters(systems)

    # collect info for tracked monster nest valid locations
    planets = [p for s in systems for p in fo.sys_get_planets(s)]
    tracked_nest_valid_locations = {nest: len(fo.special_locations(nest, planets)) for nest in nest_name_map}

    # for each system in the list that has been passed to this function, find a monster fleet that can be spawned at
    # the system and which hasn't already been added too many times, then attempt to add that monster fleet by
    # testing the spawn rate chance
    random.shuffle(systems)
    for system in systems:
        # collect info for tracked monster valid locations
        for fp in tracked_plan_valid_locations:
            if system in fp_location_cache[fp]:
                tracked_plan_valid_locations[fp] += 1

        # filter out all monster fleets whose location condition allows this system and whose counter hasn't reached 0.
        suitable_fleet_plans = [
            fp
            for fp in fleet_plans
            if system in fp_location_cache[fp]
            and spawn_limits.get(fp, 0)
            and (fp not in fleet_can_alter_starlanes or starlane_altering_monsters.can_place_at(system, fp))
        ]
        # if there are no suitable monster fleets for this system, continue with the next
        if not suitable_fleet_plans:
            continue

        # randomly select one monster fleet out of the suitable ones and then test if we want to add it to this system
        # by making a roll against the basic chance multiplied by the spawn rate of this monster fleet
        expectation_tally += (
            basic_chance * sum([fp.spawn_rate() for fp in suitable_fleet_plans]) / len(suitable_fleet_plans)
        )
        fleet_plan = random.choice(suitable_fleet_plans)
        if fleet_plan.name() in tracked_plan_tries:
            tracked_plan_tries[fleet_plan.name()] += 1
        if random.random() > basic_chance * fleet_plan.spawn_rate():
            print(
                "\t\t At system %4d rejected monster fleet %s from %d suitable fleets"
                % (system, fleet_plan.name(), len(suitable_fleet_plans))
            )
            # no, test failed, continue with the next system
            continue
        actual_tally += 1
        if fleet_plan.name() in tracked_plan_counts:
            tracked_plan_counts[fleet_plan.name()] += 1

        # all prerequisites and the test have been met, now spawn this monster fleet in this system
        # create monster fleet
        try:
            if fleet_plan in fleet_can_alter_starlanes:
                starlane_altering_monsters.place(system, fleet_plan)
            else:
                populate_monster_fleet(fleet_plan, system)
            # decrement counter for this monster fleet
            spawn_limits[fleet_plan] -= 1

        except MapGenerationError as err:
            report_error(str(err))
            continue

    print("Actual # monster fleets placed: %d; Total Placement Expectation: %.1f" % (actual_tally, expectation_tally))
    # finally, compile some statistics to be dumped to the log later
    universe_statistics.monsters_summary = [
        (fp.name(), fp.spawn_limit() - counter) for fp, counter in spawn_limits.items()
    ]
    universe_statistics.tracked_monsters_tries.update(tracked_plan_tries)
    universe_statistics.tracked_monsters_summary.update(tracked_plan_counts)
    universe_statistics.tracked_monsters_location_summary.update(
        (fp.name(), count) for fp, count in tracked_plan_valid_locations.items()
    )
    universe_statistics.tracked_nest_location_summary.update(
        (nest_name_map[nest], count) for nest, count in tracked_nest_valid_locations.items()
    )



from focs._effects import (
    Abs,
    CurrentTurn,
    EffectsGroup,
    Focus,
    IsSource,
    LocalCandidate,
    MaxOf,
    MinOf,
    Planet,
    Population,
    SetConstruction,
    SetIndustry,
    SetInfluence,
    SetResearch,
    SetStockpile,
    SetTargetConstruction,
    SetTargetIndustry,
    SetTargetInfluence,
    SetTargetResearch,
    Ship,
    Target,
    TargetPopulation,
    Value,
)
from macros.misc import DESCRIPTION_EFFECTSGROUP_MACRO
from macros.priorities import (
    AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
    END_CLEANUP_PRIORITY,
    FOCUS_CHANGE_PENALTY_PRIORITY,
    LATE_AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
)

STANDARD_METER_GROWTH = [
    # increase or decrease towards target meter of planets, when not recently conquered
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & (LocalCandidate.LastTurnConquered < CurrentTurn),
        priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=[
            SetResearch(
                value=Value
                + MinOf(
                    float,
                    1.0,  # increase by (at most) 1 / turn up towards target
                    MaxOf(float, 0.0, Value(Target.Happiness)) / 5.0,  # limit rate of increase, dependent on Stability
                    MaxOf(
                        float,
                        MinOf(
                            float,
                            -1.0,  # decrease by 1 / turn down towards target, or:
                            (Value(Target.TargetResearch) - Value)
                            / 5.0,  # when (target>=value), this is larger than -1.0.  when (target<<value) this increases the rate of meter decay
                        ),
                        Value(Target.TargetResearch)
                        - Value,  # limit increase or decrease to the difference between current and target, to avoid overshooting target
                    ),
                )
            ),
            SetIndustry(
                value=Value
                + MinOf(
                    float,
                    1.0,
                    MaxOf(float, 0.0, Value(Target.Happiness)) / 5.0,
                    MaxOf(
                        float,
                        MinOf(float, -1.0, (Value(Target.TargetIndustry) - Value) / 5.0),
                        Value(Target.TargetIndustry) - Value,
                    ),
                )
            ),
            SetInfluence(
                value=Value
                + MinOf(
                    float,
                    1.0,  # increase by (at most) 1 / turn up towards target
                    MaxOf(
                        float,
                        MinOf(
                            float,
                            -1.0,  # decrease by 1 / turn down towards target, or:
                            (Value(Target.TargetInfluence) - Value)
                            / 5.0,  # when (target>=value), this is larger than -1.0.  when (target<<value) this increases the rate of meter decay
                        ),
                        Value(Target.TargetInfluence)
                        - Value,  # limit increase or decrease to the difference between current and target, to avoid overshooting target
                    ),
                )
            ),
        ],
    ),
    # increase or decrease towards target meter of ships
    EffectsGroup(
        scope=IsSource,
        activation=Ship,
        priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=[
            SetResearch(value=Value + MinOf(float, MaxOf(float, Value(Target.TargetResearch) - Value, -1), 1)),
            SetIndustry(value=Value + MinOf(float, MaxOf(float, Value(Target.TargetIndustry) - Value, -1), 1)),
            SetInfluence(value=Value + MinOf(float, MaxOf(float, Value(Target.TargetInfluence) - Value, -1), 1)),
        ],
    ),
    # removes residual production from a dead planet
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & TargetPopulation(high=0) & Population(high=0.1),
        accountinglabel="DYING_POPULATION_LABEL",
        priority=END_CLEANUP_PRIORITY,
        effects=[
            SetResearch(value=0),
            SetIndustry(value=0),
            SetTargetIndustry(value=0),
            SetTargetResearch(value=0),
            SetTargetInfluence(value=0),
        ],
    ),
]

STANDARD_CONSTRUCTION = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet(),
        accountinglabel="STANDARD_CONSTRUCTION_LABEL",
        effects=SetTargetConstruction(value=Value + 20),
    ),
    EffectsGroup(  # increase or decrease towards target meter, when not recently conquered
        scope=IsSource,
        activation=Planet()
        & (LocalCandidate.LastTurnConquered < CurrentTurn)
        & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn),
        priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=SetConstruction(
            value=Value + MinOf(float, MaxOf(float, Value(Target.TargetConstruction) - Value, -1), 1)
        ),
    ),
    EffectsGroup(  # always ensure minimum value of one, as this is necessary for being attacked
        scope=IsSource,
        activation=Planet(),
        # has to happen after e.g. FORCE_ENERGY_STRC effects which also happens at AFTER_ALL_TARGET_MAX_METERS_PRIORITY
        priority=LATE_AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=SetConstruction(value=MaxOf(float, Value, 1)),
    ),
]

FOCUS_CHANGE_PENALTY = [
    # penalize changing focus by reducing the non-focused resource meters on the turns after the focus change.
    # these effects reduce the non-focused resource meters by 2 on the turn after a focus change.
    # eg. if changing to industry focus, if the new research target is 10 and the research is 20, research will
    # be reduced by 2 to 18. If research is 11, then it will be only reduced to the target (10).
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & ~Focus(type=["FOCUS_INDUSTRY"]) & (0 == LocalCandidate.TurnsSinceFocusChange),
        priority=FOCUS_CHANGE_PENALTY_PRIORITY,
        effects=SetIndustry(
            value=Value
            - MaxOf(
                float,
                0,
                MinOf(
                    float,
                    MaxOf(float, 2, (Abs(float, Value - Value(Target.TargetIndustry))) ** 0.5),
                    Value - Value(Target.TargetIndustry),
                ),
            )
        ),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & ~Focus(type=["FOCUS_RESEARCH"]) & (0 == LocalCandidate.TurnsSinceFocusChange),
        priority=FOCUS_CHANGE_PENALTY_PRIORITY,
        effects=SetResearch(
            value=Value
            - MaxOf(
                float,
                0,
                MinOf(
                    float,
                    MaxOf(float, 2, (Abs(float, Value - Value(Target.TargetResearch))) ** 0.5),
                    Value - Value(Target.TargetResearch),
                ),
            )
        ),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & ~Focus(type=["FOCUS_INFLUENCE"]) & (0 == LocalCandidate.TurnsSinceFocusChange),
        priority=FOCUS_CHANGE_PENALTY_PRIORITY,
        effects=SetInfluence(
            value=Value
            - MaxOf(
                float,
                0,
                MinOf(
                    float,
                    MaxOf(float, 2, (Abs(float, Value - Value(Target.TargetInfluence))) ** 0.5),
                    Value - Value(Target.TargetInfluence),
                ),
            )
        ),
    ),
    # TODO Delete this whole Stockpile penalty clause if Stockpile is determined to remain as a Max rather than a Target type meter
    # If a decision is made to change Stockpile to a Target type meter, then simply change MaxStockpile below to TargetStockpile
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & ~Focus(type=["FOCUS_STOCKPILE"]) & (0 == LocalCandidate.TurnsSinceFocusChange),
        priority=FOCUS_CHANGE_PENALTY_PRIORITY,
        effects=SetStockpile(
            value=Value
            - MaxOf(float, 0, MinOf(float, 1.0 - Target.TurnsSinceFocusChange, Value - Value(Target.MaxStockpile)))
        ),
    ),
]


# FAST_COLONIZATION and SLOW_COLONIZATION are stubs for applying the
# fast colonization building speed description to a species.
FAST_COLONIZATION = DESCRIPTION_EFFECTSGROUP_MACRO("FAST_COLONIZATION_DESC")
SLOW_COLONIZATION = DESCRIPTION_EFFECTSGROUP_MACRO("SLOW_COLONIZATION_DESC")



from abc import ABC, abstractmethod
from typing import Any


class Field(ABC):
    def __init__(
        self,
        name: str,
        align: str = "<",
        description: str = "",
        placeholder: str = "",
        total: bool = False,
    ):
        """
        Cell specification.
        :param name: name of the column
        :param align: "<" | ">" | "=" | "^" https://docs.python.org/2/library/string.html#format-specification-mini-language
        :param description: description for column name, will be printed in table legend,
                            specify it if you use abbr as name: ``name``="PP", ``description``="production points"
        :param placeholder: placeholder for empty value
        :param total: add total string
        """
        self.name = name
        self.align = align
        self.description = description
        self.placeholder = placeholder
        if description:
            self.name += "*"
        self.total = total

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"

    def format_cell(self, item, width):
        return f"{item:{self.align}{width}}"

    def format_header(self, width):
        return f"{self.name:<{width}}"

    def make_cell_string(self, val):
        if self.placeholder and not val:
            return self.placeholder
        else:
            return self.convert_value_to_string(val)

    def make_summary_value(self):
        return ""

    @abstractmethod
    def convert_value_to_string(self, val: Any) -> str:
        ...



import pytest

from character.character_module import Character, Trait


class LeftTrait(Trait):
    """A test trait"""


class RightTrait(Trait):
    """A test Trait that injects values to be found by combiners"""

    def may_explore_system(self, monster_threat):  # pylint: disable=no-self-use,unused-argument
        """Always reject. 1 parameter"""
        return False

    def may_maximize_research(self):  # pylint: disable=no-self-use,unused-argument
        """Always reject. 0 parameters"""
        return False

    def preferred_research_cutoff(self, alternatives):  # pylint: disable=no-self-use,unused-argument
        """Pick 1st"""
        return alternatives[1]

    def max_number_colonies(self):  # pylint: disable=no-self-use,unused-argument
        """Use for min combiner"""
        return 100

    def military_safety_factor(self):  # pylint: disable=no-self-use,unused-argument
        """Use for max combiner"""
        return 10


class OtherTrait(Trait):
    """A test trait"""


left_trait = LeftTrait()
right_trait = RightTrait()
other_trait = OtherTrait()

rejection_character = Character([left_trait, right_trait])
permissive_character = Character([left_trait, other_trait])


class TestCharacter:
    """Test the Character class which combines traits
    Each combiner test checks that the combiner generates the expected output.
    """

    def test_get_trait(self):
        assert rejection_character.get_trait(LeftTrait) == left_trait
        assert rejection_character.get_trait(RightTrait) == right_trait
        assert rejection_character.get_trait(OtherTrait) != left_trait
        assert rejection_character.get_trait(OtherTrait) != right_trait
        assert rejection_character.get_trait(OtherTrait) is None

    def test_all_combiner(self):
        assert permissive_character.may_maximize_research()
        assert not rejection_character.may_maximize_research()

    def test_all_combiner_one_parameter(self):
        assert permissive_character.may_explore_system(0)
        assert not rejection_character.may_explore_system(0)

    def test_min_combiner(self):
        assert permissive_character.max_number_colonies() == 1000000
        assert rejection_character.max_number_colonies() == 100

    def test_max_combiner(self):
        assert permissive_character.military_safety_factor() == 0
        assert rejection_character.military_safety_factor() == 10

    def test_preference_combiner(self):
        assert permissive_character.preferred_research_cutoff([10, 11, 12]) is None
        assert rejection_character.preferred_research_cutoff([10, 11, 12]) == 11

    def test_character_must_be_composed_of_traits(self):
        with pytest.raises(TypeError, match="All traits must be sub-classes of Trait"):
            not_a_trait = 1
            Character([LeftTrait, not_a_trait])



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_MINIAST_SWARM",
    description="SHP_MINIAST_SWARM_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=128 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_ASTEROID_HULL_TECHS"],
    prerequisites=["SHP_ASTEROID_REFORM"],
    unlock=[
        Item(type=UnlockShipHull, name="SH_MINIASTEROID_SWARM"),
    ],
    graphic="icons/ship_hulls/mini_asteroid_swarm_small.png",
)



from focs._effects import (
    EffectsGroup,
    OwnedBy,
    Planet,
    SetMaxSupply,
    Source,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="CON_ORBITAL_CON",
    description="CON_ORBITAL_CON_DESC",
    short_description="SUPPLY_SHORT_DESC",
    category="CONSTRUCTION_CATEGORY",
    researchcost=72 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_CONSTRUCTION_CATEGORY"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            accountinglabel="CON_TECH_ACCOUNTING_LABEL",
            effects=SetMaxSupply(value=Value + 1),
        )
    ],
    graphic="icons/tech/orbital_construction.png",
)



VERY_BAD_MULTIPLIER = 0.5

BAD_MULTIPLIER = 0.75

AVERAGE_MULTIPLIER = 1.0

GOOD_MULTIPLIER = 1.5

GREAT_MULTIPLIER = 2.0

ULTIMATE_MULTIPLIER = 3.0

BAD_MULTIPLIER_WHOLE = 0

AVERAGE_MULTIPLIER_WHOLE = 1

GOOD_MULTIPLIER_WHOLE = 2

GREAT_MULTIPLIER_WHOLE = 3

ULTIMATE_MULTIPLIER_WHOLE = 4

ADEQUATE_MULTIPLIER = 0.75

POOR_MULTIPLIER = 0.5

HOSTILE_MULTIPLIER = 0.0
